"""
ElevenLabs STT provider with OpenAI-compatible interface.

This module provides an OpenAI Whisper API-compatible interface for ElevenLabs
speech-to-text transcription service.
"""

from pathlib import Path
from typing import Any, BinaryIO, Generator, List, Optional, Union, cast

import requests

from webscout import exceptions
from webscout.litagent import LitAgent
from webscout.Provider.STT.base import (
    BaseSTTAudio,
    BaseSTTTranscriptions,
    STTCompatibleProvider,
    STTModels,
    TranscriptionResponse,
)


class ElevenLabsTranscriptions(BaseSTTTranscriptions):
    """ElevenLabs transcriptions interface."""

    def create(
        self,
        *,
        model: str,
        file: Union[BinaryIO, str, Path],
        language: Optional[str] = None,
        prompt: Optional[str] = None,
        response_format: str = "json",
        temperature: Optional[float] = None,
        timestamp_granularities: Optional[List[str]] = None,
        stream: bool = False,
        timeout: Optional[int] = None,
        proxies: Optional[dict] = None,
        **kwargs: Any,
    ) -> Union[TranscriptionResponse, Generator[str, None, None]]:
        """Create a transcription using ElevenLabs API."""
        # Always use file as file-like object
        if isinstance(file, (str, Path)):
            audio_file = open(str(file), "rb")
            close_file = True
        else:
            audio_file = file
            close_file = False
        try:
            if stream:
                return self._create_stream(
                    audio_file=audio_file,
                    model=model,
                    language=language,
                    prompt=prompt,
                    response_format=response_format,
                    temperature=temperature,
                    timestamp_granularities=timestamp_granularities,
                    timeout=timeout,
                    proxies=proxies,
                    **kwargs,
                )
            else:
                result = self._create_non_stream(
                    audio_file=audio_file,
                    model=model,
                    language=language,
                    prompt=prompt,
                    response_format=response_format,
                    temperature=temperature,
                    timestamp_granularities=timestamp_granularities,
                    timeout=timeout,
                    proxies=proxies,
                    **kwargs,
                )
                return result
        finally:
            if close_file:
                audio_file.close()

    def _create_non_stream(
        self,
        audio_file: BinaryIO,
        model: str,
        language: Optional[str] = None,
        prompt: Optional[str] = None,
        response_format: str = "json",
        temperature: Optional[float] = None,
        timestamp_granularities: Optional[List[str]] = None,
        timeout: Optional[int] = None,
        proxies: Optional[dict] = None,
        **kwargs: Any,
    ) -> TranscriptionResponse:
        """Create non-streaming transcription."""
        try:
            headers = {
                "Accept": "application/json, text/plain, */*",
                "Accept-Language": "en-US,en;q=0.9",
                "User-Agent": LitAgent().random(),
            }
            api_url = self._client.api_url
            if getattr(self._client, "allow_unauthenticated", False):
                if "?" in api_url:
                    api_url += "&allow_unauthenticated=1"
                else:
                    api_url += "?allow_unauthenticated=1"
            files = {
                "file": audio_file,
                "model_id": (None, self._client.model_id),
                "tag_audio_events": (None, "true" if self._client.tag_audio_events else "false"),
                "diarize": (None, "true" if self._client.diarize else "false"),
            }
            if language:
                files["language"] = (None, language)
            response = requests.post(
                api_url,
                files=files,
                headers=headers,
                timeout=timeout or self._client.timeout,
                proxies=proxies or getattr(self._client, "proxies", None),
            )
            if response.status_code != 200:
                raise exceptions.FailedToGenerateResponseError(
                    f"ElevenLabs API returned error: {response.status_code} - {response.text}"
                )
            result = response.json()
            simple_result = {"text": result.get("text", "")}
            return TranscriptionResponse(simple_result, response_format)
        except Exception as e:
            raise exceptions.FailedToGenerateResponseError(
                f"ElevenLabs transcription failed: {str(e)}"
            )

    def _create_stream(
        self,
        audio_file: BinaryIO,
        model: str,
        language: Optional[str] = None,
        prompt: Optional[str] = None,
        response_format: str = "json",
        temperature: Optional[float] = None,
        timestamp_granularities: Optional[List[str]] = None,
        timeout: Optional[int] = None,
        proxies: Optional[dict] = None,
        **kwargs: Any,
    ) -> Generator[str, None, None]:
        """Create streaming transcription using requests.post(..., stream=True)."""
        headers = {
            "Accept": "application/json, text/plain, */*",
            "Accept-Language": "en-US,en;q=0.9",
            "User-Agent": LitAgent().random(),
        }
        api_url = self._client.api_url
        if getattr(self._client, "allow_unauthenticated", False):
            if "?" in api_url:
                api_url += "&allow_unauthenticated=1"
            else:
                api_url += "?allow_unauthenticated=1"
        files = {
            "file": audio_file,
            "model_id": (None, self._client.model_id),
            "tag_audio_events": (None, "true" if self._client.tag_audio_events else "false"),
            "diarize": (None, "true" if self._client.diarize else "false"),
        }
        if language:
            files["language"] = (None, language)
        response = requests.post(
            api_url,
            files=files,
            headers=headers,
            timeout=timeout or self._client.timeout,
            proxies=proxies or getattr(self._client, "proxies", None),
            stream=True,
        )
        if response.status_code != 200:
            raise exceptions.FailedToGenerateResponseError(
                f"ElevenLabs API returned error: {response.status_code} - {response.text}"
            )
        # Stream the response, decode utf-8
        for line in response.iter_lines(decode_unicode=True):
            if line:
                yield line


class ElevenLabsAudio(BaseSTTAudio):
    """ElevenLabs audio interface."""

    def _create_transcriptions(self, client) -> ElevenLabsTranscriptions:
        return ElevenLabsTranscriptions(client)


class ElevenLabsSTT(STTCompatibleProvider):
    """
    OpenAI-compatible client for ElevenLabs STT API.

    Usage:
        client = ElevenLabsSTT()
        audio_file = open("audio.mp3", "rb")
        transcription = client.audio.transcriptions.create(
            model="scribe_v1",
            file=audio_file,
            response_format="text"
        )
        print(transcription.text)
    """

    AVAILABLE_MODELS = [
        "scribe_v1",
    ]

    def __init__(
        self,
        model_id: str = "scribe_v1",
        allow_unauthenticated: bool = True,
        tag_audio_events: bool = True,
        diarize: bool = True,
        timeout: int = 60,
        proxies: Optional[dict] = None,
    ):
        """Initialize ElevenLabs STT provider."""
        self.model_id = model_id
        self.allow_unauthenticated = allow_unauthenticated
        self.tag_audio_events = tag_audio_events
        self.diarize = diarize
        self.timeout = timeout
        self.proxies = proxies

        # API configuration
        self.api_url = "https://api.elevenlabs.io/v1/speech-to-text"

        # Initialize interfaces
        self.audio = ElevenLabsAudio(self)
        self._models = STTModels(self.AVAILABLE_MODELS)

    @property
    def models(self):
        """Get models interface."""
        return self._models


if __name__ == "__main__":
    from rich import print

    client = ElevenLabsSTT()

    # Example audio file path - replace with your own
    audio_file_path = r"C:\Users\koula\Downloads\audio_2025-05-12_22-30-47.ogg"

    print("=== Non-streaming example ===")
    with open(audio_file_path, "rb") as audio_file:
        transcription = client.audio.transcriptions.create(
            model="scribe_v1", file=audio_file, stream=False
        )
        if hasattr(transcription, 'text'):
            print(transcription.text)

    print("\n=== Streaming example ===")
    with open(audio_file_path, "rb") as audio_file:
        stream_gen = client.audio.transcriptions.create(
            model="scribe_v1", file=audio_file, stream=True
        )
        for chunk in cast(Generator[str, None, None], stream_gen):
            print(chunk.strip())



"""Command decorators for SwiftCLI."""

from functools import wraps
from typing import Any, Callable, Dict, List, Optional

from ..core.context import Context


def command(
    name: Optional[str] = None,
    help: Optional[str] = None,
    aliases: Optional[List[str]] = None,
    hidden: bool = False,
) -> Callable:
    """
    Decorator to register a new command.

    This decorator marks a function as a CLI command and provides metadata
    about how command should be registered and displayed.

    Args:
        name: Command name (defaults to function name)
        help: Help text (defaults to function docstring)
        aliases: Alternative names for command
        hidden: Whether to hide from help output

    Example:
        @command(name="greet", help="Say hello")
        def hello(name: str):
            print(f"Hello {name}!")

        @command(aliases=["hi", "hey"])
        def hello(name: str):
            '''Say hello to someone'''
            print(f"Hello {name}!")
    """

    def decorator(f: Callable) -> Callable:
        f._command = {  # type: ignore[attr-defined]
            "name": name or getattr(f, '__name__', 'unknown'),
            "help": help or f.__doc__,
            "aliases": aliases or [],
            "hidden": hidden,
        }
        return f

    return decorator


def group(
    name: Optional[str] = None,
    help: Optional[str] = None,
    chain: bool = False,
    invoke_without_command: bool = False,
) -> Callable:
    """
    Decorator to create a command group.

    Command groups can contain subcommands and optionally chain their results.

    Args:
        name: Group name (defaults to function name)
        help: Help text (defaults to function docstring)
        chain: Whether to chain command results
        invoke_without_command: Allow group to be invoked without subcommand

    Example:
        @group()
        def db():
            '''Database commands'''
            pass

        @db.command()
        def migrate():
            '''Run database migrations'''
            print("Running migrations...")

        @group(chain=True)
        def process():
            '''Process data'''
            pass

        @process.command()
        def validate():
            '''Validate data'''
            return {"valid": True}
    """

    def decorator(f: Callable) -> Callable:
        f._group = {  # type: ignore[attr-defined]
            "name": name or getattr(f, '__name__', 'unknown'),
            "help": help or f.__doc__,
            "chain": chain,
            "invoke_without_command": invoke_without_command,
        }
        return f

    return decorator


def pass_context(f: Callable) -> Callable:
    """
    Decorator to pass CLI context to command.

    This decorator injects current Context object as first argument
    to the decorated command function.

    Example:
        @command()
        @pass_context
        def status(ctx):
            '''Show application status'''
            print(f"App: {ctx.cli.name}")
            print(f"Debug: {ctx.debug}")
    """
    f._pass_context = True  # type: ignore[attr-defined]
    return f


def completion(func: Optional[Callable] = None) -> Callable:
    """
    Decorator to provide shell completion for a command.

    The decorated function should return a list of possible completions
    based on current incomplete value.

    Example:
        @command()
        @option("--service", type=str)
        def restart(service: str):
            '''Restart a service'''
            print(f"Restarting {service}...")

        @restart.completion()
        def complete_service(ctx, incomplete):
            services = ["nginx", "apache", "mysql"]
            return [s for s in services if s.startswith(incomplete)]
    """

    def decorator(f: Callable) -> Callable:
        @wraps(f)
        def wrapper(ctx: Context, incomplete: str) -> List[str]:
            try:
                return f(ctx, incomplete)
            except Exception:
                return []

        return wrapper

    if func is None:
        return decorator
    return decorator(func)


def argument(
    name: str,
    type: Any = str,
    required: bool = True,
    help: Optional[str] = None,
    default: Any = None,
    validation: Optional[Dict[str, Any]] = None,
    mutually_exclusive: Optional[List[str]] = None,
) -> Callable:
    """
    Decorator to add a command argument.

    Arguments are positional parameters that must be provided in order.

    Args:
        name: Argument name
        type: Expected type
        required: Whether argument is required
        help: Help text
        default: Default value if not required
        validation: Dictionary of validation rules (min_length, max_length, pattern, choices, etc.)
        mutually_exclusive: List of argument names that are mutually exclusive with this argument

    Example:
        @command()
        @argument("name", validation={'min_length': 2, 'max_length': 50})
        @argument("count", type=int, default=1, validation={'min': 1, 'max': 100})
        def greet(name: str, count: int):
            '''Greet someone multiple times'''
            for _ in range(count):
                print(f"Hello {name}!")
    """

    def decorator(f: Callable) -> Callable:
        if not hasattr(f, "_arguments"):
            f._arguments = []  # type: ignore[attr-defined]

        f._arguments.append(  # type: ignore[attr-defined]
            {
                "name": name,
                "type": type,
                "required": required,
                "help": help,
                "default": default,
                "validation": validation,
                "mutually_exclusive": mutually_exclusive,
            }
        )
        return f

    return decorator


def flag(name: str, help: Optional[str] = None, hidden: bool = False) -> Callable:
    """
    Decorator to add a boolean flag option.

    Flags are special options that don't take a value - they're either
    present (True) or absent (False).

    Args:
        name: Flag name
        help: Help text
        hidden: Whether to hide from help output

    Example:
        @command()
        @flag("--verbose", help="Enable verbose output")
        def process(verbose: bool):
            '''Process data'''
            if verbose:
                print("Verbose mode enabled")
    """

    def decorator(f: Callable) -> Callable:
        if not hasattr(f, "_options"):
            f._options = []  # type: ignore[attr-defined]

        f._options.append(  # type: ignore[attr-defined]
            {
                "param_decls": [name],
                "is_flag": True,
                "help": help,
                "hidden": hidden,
            }
        )
        return f

    return decorator



import json
from typing import Any, Dict, Generator, Optional, Union, cast

import requests

from webscout import exceptions
from webscout.AIbase import Provider
from webscout.AIutel import AwesomePrompts, Conversation, Optimizers, sanitize_stream


class ChatHub(Provider):
    """
    A class to interact with the ChatHub API.
    """

    AVAILABLE_MODELS = [
        'meta/llama3.1-8b',
        'mistral/mixtral-8x7b',
        'google/gemma-2',
        'perplexity/sonar-online',
    ]
    model_aliases = {  # Aliases for shorter model names
        "llama3.1-8b": 'meta/llama3.1-8b',
        "mixtral-8x7b": 'mistral/mixtral-8x7b',
        "gemma-2": 'google/gemma-2',
        "sonar-online": 'perplexity/sonar-online',
    }
    default_model = "sonar-online"


    def __init__(
        self,
        is_conversation: bool = True,
        max_tokens: int = 2049,
        timeout: int = 30,
        intro: Optional[str] = None,
        filepath: Optional[str] = None,
        update_file: bool = True,
        proxies: dict = {},
        history_offset: int = 10250,
        act: Optional[str] = None,
        model: str = "sonar-online",
    ):
        """Initializes the ChatHub API client."""
        self.url = "https://app.chathub.gg"
        self.api_endpoint = "https://app.chathub.gg/api/v3/chat/completions"
        self.headers = {
            'Accept': '*/*',
            'Accept-Language': 'en-US,en;q=0.9',
            'Content-Type': 'application/json',
            'Origin': self.url,
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/129.0.0.0 Safari/537.36',
            'X-App-Id': 'web'
        }
        self.session = requests.Session()
        self.session.headers.update(self.headers)
        self.session.proxies.update(proxies)
        self.timeout = timeout
        self.last_response = {}

        self.is_conversation = is_conversation
        self.max_tokens_to_sample = max_tokens
        self.__available_optimizers = (
            method
            for method in dir(Optimizers)
            if callable(getattr(Optimizers, method)) and not method.startswith("__")
        )
        self.conversation = Conversation(
            is_conversation, self.max_tokens_to_sample, filepath, update_file
        )
        act_prompt = (
            AwesomePrompts().get_act(cast(Union[str, int], act), default=None, case_insensitive=True
            )
            if act
            else intro
        )
        if act_prompt:
            self.conversation.intro = act_prompt
        self.conversation.history_offset = history_offset

        #Resolve the model
        self.model = self.get_model(model)


    def get_model(self, model: str) -> str:
        """
        Resolves the model name using aliases or defaults.
        """

        if model in self.AVAILABLE_MODELS:
            return model
        elif model in self.model_aliases:
            return self.model_aliases[model]
        else:
            print(f"Model '{model}' not found. Using default model '{self.default_model}'.")
            return self.default_model # Use class-level default

    def ask(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False,
        optimizer: Optional[str] = None,
        conversationally: bool = False,
        **kwargs: Any,
    ) -> Union[Dict[str, Any], Generator]:

        conversation_prompt = self.conversation.gen_complete_prompt(prompt)
        if optimizer:
            if optimizer in self.__available_optimizers:
                conversation_prompt = getattr(Optimizers, optimizer)(
                    conversation_prompt if conversationally else prompt
                )
            else:
                raise Exception(f"Optimizer is not one of {self.__available_optimizers}")


        data = {
            "model": self.model,
            "messages": [{"role": "user", "content": conversation_prompt}],
            "tools": []
        }

        # Set the Referer header dynamically based on the resolved model
        self.headers['Referer'] = f"{self.url}/chat/{self.model}"


        def for_stream():
            try:
                with requests.post(self.api_endpoint, headers=self.headers, json=data, stream=True, timeout=self.timeout) as response:
                    response.raise_for_status()
                    streaming_text = ""

                    # Use sanitize_stream for processing
                    processed_stream = sanitize_stream(
                        data=response.iter_lines(decode_unicode=True),
                        intro_value="data:",
                        to_json=True,
                        content_extractor=lambda x: x.get('textDelta') if isinstance(x, dict) else None,
                        skip_markers=["[DONE]"],
                        yield_raw_on_error=False,
                        raw=raw
                    )

                    for content in processed_stream:
                        if content:
                            if raw:
                                yield content if isinstance(content, str) else json.dumps(content)
                            else:
                                streaming_text += content if isinstance(content, str) else ""
                                yield dict(text=content)

                    self.conversation.update_chat_history(prompt, streaming_text)
                    self.last_response = {"text": streaming_text}
            except requests.exceptions.RequestException as e:
                raise exceptions.FailedToGenerateResponseError(f"Request error: {e}")


        def for_non_stream():
            for _ in for_stream():
                pass
            return self.last_response if not raw else json.dumps(self.last_response)

        return for_stream() if stream else for_non_stream()




    def chat(
        self,
        prompt: str,
        stream: bool = False,
        optimizer: Optional[str] = None,
        conversationally: bool = False,
        **kwargs: Any,
    ) -> Union[str, Generator[str, None, None]]:
        """Generate response `str`"""

        def for_stream():
            for response in self.ask(
                prompt, stream=True, raw=False, optimizer=optimizer, conversationally=conversationally
            ):
                yield self.get_message(response)

        def for_non_stream():
            result = self.ask(
                prompt,
                stream=False,
                raw=False,
                optimizer=optimizer,
                conversationally=conversationally,
            )
            return self.get_message(result)

        return for_stream() if stream else for_non_stream()



    def get_message(self, response: Union[Dict[str, Any], Generator[Any, None, None], str]) -> str:
        """Retrieves message only from response"""
        if isinstance(response, dict):
            resp_dict = cast(Dict[str, Any], response)
            return resp_dict.get("text", "")
        elif isinstance(response, str):
            return response
        else:
            # Generator case - shouldn't happen here
            return ""


if __name__ == "__main__":
    from rich import print
    bot = ChatHub()
    try:
        response = bot.chat("tell me about Abhay koul, HelpingAI", stream=True)
        for chunk in response:
            print(chunk, end="", flush=True)
    except Exception as e:
        print(f"An error occurred: {e}")



import os
import platform
import signal
import subprocess
import sys
import tempfile
from pathlib import Path
from typing import Dict, List, Optional, Set, TypedDict

from rich.console import Console
from rich.panel import Panel
from rich.table import Table

from webscout.swiftcli import CLI, option
from webscout.zeroart import figlet_format

# Lazy import for huggingface_hub
HfApi = None

def _ensure_huggingface_hub():
    """Ensure huggingface_hub is imported when needed."""
    global HfApi
    if HfApi is None:
        try:
            from huggingface_hub import HfApi as _HfApi  # type: ignore
            HfApi = _HfApi
        except ImportError:
            raise ImportError(
                "huggingface_hub is required for GGUF conversion. "
                "Install it with: pip install huggingface_hub"
            )
    return HfApi

console = Console()

class ConversionError(Exception):
    """Custom exception for when things don't go as planned! ⚠️"""
    pass

class QuantizationMethod(TypedDict):
    """Type definition for quantization method descriptions."""
    description: str

class ModelConverter:
    """Handles the conversion of Hugging Face models to GGUF format."""

    VALID_METHODS: Dict[str, str] = {
        # Full precision types
        "f32": "32-bit floating point - full precision, largest size",
        "fp16": "16-bit floating point - maximum accuracy, large size",
        "f16": "16-bit floating point - alias for fp16",
        "bf16": "bfloat16 - good balance for training and some models",
        "auto": "Auto-detect best 16-bit type based on model tensors",
        # K-Quant types (super-block quantization)
        "q2_k": "2-bit K-quant (smallest size, lowest accuracy)",
        "q2_k_s": "2-bit K-quant small - maximum compression",
        "q3_k_l": "3-bit K-quant large - balanced for size/accuracy",
        "q3_k_m": "3-bit K-quant medium - good balance for most use cases",
        "q3_k_s": "3-bit K-quant small - optimized for speed",
        "q4_0": "4-bit quantization (legacy) - standard 4-bit, auto-repacks for ARM",
        "q4_1": "4-bit quantization (legacy) - improved accuracy over q4_0",
        "q4_k_l": "4-bit K-quant large - highest quality 4-bit",
        "q4_k_m": "4-bit K-quant medium - balanced for most models",
        "q4_k_s": "4-bit K-quant small - optimized for speed",
        "q5_0": "5-bit quantization (legacy) - high accuracy",
        "q5_1": "5-bit quantization (legacy) - improved accuracy over q5_0",
        "q5_k_l": "5-bit K-quant large - highest quality 5-bit",
        "q5_k_m": "5-bit K-quant medium - best balance for quality/size",
        "q5_k_s": "5-bit K-quant small - optimized for speed",
        "q6_k": "6-bit K-quant - near-lossless quality",
        "q8_0": "8-bit quantization - near-original quality",
        # Ternary quantization (experimental)
        "tq1_0": "1-bit ternary quantization - experimental, extreme compression",
        "tq2_0": "2-bit ternary quantization - experimental, very small size"
    }

    VALID_IMATRIX_METHODS: Dict[str, str] = {
        # 1-bit IQ types (extreme compression, requires imatrix)
        "iq1_s": "1-bit IQ small - extreme compression, requires imatrix",
        "iq1_m": "1-bit IQ medium - extreme compression, requires imatrix",
        # 2-bit IQ types
        "iq2_xxs": "2-bit IQ extra extra small - maximum compression",
        "iq2_xs": "2-bit IQ extra small - very high compression",
        "iq2_s": "2-bit IQ small - high compression",
        "iq2_m": "2-bit IQ medium - balanced compression",
        # 3-bit IQ types
        "iq3_xxs": "3-bit IQ extra extra small - maximum compression",
        "iq3_xs": "3-bit IQ extra small - high compression",
        "iq3_s": "3-bit IQ small - balanced compression",
        "iq3_m": "3-bit IQ medium - balanced importance-based",
        # 4-bit IQ types
        "iq4_nl": "4-bit IQ non-linear - best accuracy for 4-bit",
        "iq4_xs": "4-bit IQ extra small - maximum 4-bit compression",
        # K-Quant types that benefit from imatrix
        "q2_k": "2-bit K-quant with imatrix - improved quality",
        "q2_k_s": "2-bit K-quant small with imatrix",
        "q3_k_s": "3-bit K-quant small with imatrix",
        "q3_k_m": "3-bit K-quant medium with imatrix",
        "q3_k_l": "3-bit K-quant large with imatrix",
        "q4_k_s": "4-bit K-quant small with imatrix",
        "q4_k_m": "4-bit K-quant medium with imatrix",
        "q5_k_s": "5-bit K-quant small with imatrix",
        "q5_k_m": "5-bit K-quant medium with imatrix",
        "q6_k": "6-bit K-quant with imatrix"
    }
    # Default output type options (matches llama.cpp)
    VALID_OUTTYPES: Set[str] = {"f32", "f16", "bf16", "q8_0", "tq1_0", "tq2_0", "auto"}

    def __init__(
        self,
        model_id: str,
        username: Optional[str] = None,
        token: Optional[str] = None,
        quantization_methods: str = "q4_k_m",
        use_imatrix: bool = False,
        train_data_file: Optional[str] = None,
        split_model: bool = False,
        split_max_tensors: int = 256,
        split_max_size: Optional[str] = None,
        # New llama.cpp options
        outtype: str = "f16",
        vocab_only: bool = False,
        remote: bool = False,
        dry_run: bool = False,
        no_lazy: bool = False,
        model_name: Optional[str] = None,
        small_first_shard: bool = False
    ) -> None:
        self.model_id = model_id
        self.username = username
        self.token = token
        self.quantization_methods = quantization_methods.split(',')
        self.model_name = model_name or model_id.split('/')[-1]
        self.workspace = Path(os.getcwd())
        self.use_imatrix = use_imatrix
        self.train_data_file = train_data_file
        self.use_split = split_model
        self.split_max_tensors = split_max_tensors
        self.split_max_size = split_max_size
        # New llama.cpp options
        self.outtype = outtype.lower()
        self.vocab_only = vocab_only
        self.remote = remote
        self.dry_run = dry_run
        self.no_lazy = no_lazy
        self.small_first_shard = small_first_shard
        # Determine if we only need the base conversion (no quantization)
        self.base_only = self.outtype in self.VALID_OUTTYPES and len(self.quantization_methods) == 1 and self.quantization_methods[0] in ["fp16", "f16", "f32", "bf16", "auto"]

    def validate_inputs(self) -> None:
        """Validates all input parameters."""
        if '/' not in self.model_id:
            raise ValueError("Invalid model ID format. Expected format: 'organization/model-name'")

        if self.use_imatrix:
            invalid_methods = [m for m in self.quantization_methods if m not in self.VALID_IMATRIX_METHODS]
            if invalid_methods:
                raise ValueError(
                    f"Invalid imatrix quantization methods: {', '.join(invalid_methods)}.\n"
                    f"Valid methods are: {', '.join(self.VALID_IMATRIX_METHODS.keys())}"
                )
            if not self.train_data_file and not os.path.exists("llama.cpp/groups_merged.txt"):
                raise ValueError("Training data file is required for imatrix quantization")
        else:
            invalid_methods = [m for m in self.quantization_methods if m not in self.VALID_METHODS]
            if invalid_methods:
                raise ValueError(
                    f"Invalid quantization methods: {', '.join(invalid_methods)}.\n"
                    f"Valid methods are: {', '.join(self.VALID_METHODS.keys())}"
                )

        if bool(self.username) != bool(self.token):
            raise ValueError("Both username and token must be provided for upload, or neither.")

        # Validate outtype
        if self.outtype not in self.VALID_OUTTYPES:
            raise ValueError(
                f"Invalid output type: {self.outtype}.\n"
                f"Valid types are: {', '.join(self.VALID_OUTTYPES)}"
            )

        if self.use_split and self.split_max_size:
            try:
                # Support K, M, G units (like llama.cpp's split_str_to_n_bytes)
                if self.split_max_size[-1].upper() in ['K', 'M', 'G']:
                    int(self.split_max_size[:-1])
                    unit = self.split_max_size[-1].upper()
                    if unit not in ['K', 'M', 'G']:
                        raise ValueError("Split max size must end with K, M, or G")
                elif self.split_max_size.isnumeric():
                    int(self.split_max_size)
                else:
                    raise ValueError("Invalid format")
            except (ValueError, IndexError):
                raise ValueError("Invalid split max size format. Use format like '256M', '5G', or numeric bytes")

    @staticmethod
    def check_dependencies() -> Dict[str, bool]:
        """Check if all required dependencies are installed with cross-platform support."""
        system = platform.system()

        dependencies: Dict[str, str] = {
            'git': 'Git version control',
            'cmake': 'CMake build system',
            'ninja': 'Ninja build system (optional)'
        }

        # Add platform-specific dependencies
        if system != 'Windows':
            dependencies['pip3'] = 'Python package installer'
        else:
            dependencies['pip'] = 'Python package installer'

        status: Dict[str, bool] = {}

        for cmd, desc in dependencies.items():
            try:
                if system == 'Windows':
                    # Use 'where' command on Windows
                    result = subprocess.run(['where', cmd], capture_output=True, text=True)
                    status[cmd] = result.returncode == 0
                else:
                    # Use 'which' command on Unix-like systems
                    result = subprocess.run(['which', cmd], capture_output=True, text=True)
                    status[cmd] = result.returncode == 0
            except (FileNotFoundError, subprocess.SubprocessError):
                status[cmd] = False

        # Special check for Python - try different variants
        python_variants = ['python3', 'python', 'py'] if system != 'Windows' else ['python', 'py', 'python3']
        status['python'] = False
        for variant in python_variants:
            try:
                if system == 'Windows':
                    result = subprocess.run(['where', variant], capture_output=True)
                else:
                    result = subprocess.run(['which', variant], capture_output=True)
                if result.returncode == 0:
                    status['python'] = True
                    break
            except Exception:
                continue

        # Check for C++ compiler
        cpp_compilers = ['cl', 'g++', 'clang++'] if system == 'Windows' else ['g++', 'clang++']
        status['cpp_compiler'] = False
        for compiler in cpp_compilers:
            try:
                if system == 'Windows':
                    result = subprocess.run(['where', compiler], capture_output=True)
                else:
                    result = subprocess.run(['which', compiler], capture_output=True)
                if result.returncode == 0:
                    status['cpp_compiler'] = True
                    break
            except Exception:
                continue

        dependencies['python'] = 'Python interpreter'
        dependencies['cpp_compiler'] = 'C++ compiler (g++, clang++, or MSVC)'

        return status

    def detect_hardware(self) -> Dict[str, bool]:
        """Detect available hardware acceleration with improved cross-platform support."""
        hardware: Dict[str, bool] = {
            'cuda': False,
            'metal': False,
            'opencl': False,
            'vulkan': False,
            'rocm': False,
            'blas': False,
            'accelerate': False
        }

        system = platform.system()

        # Check CUDA
        try:
            # Check for nvcc compiler
            if subprocess.run(['nvcc', '--version'], capture_output=True, shell=(system == 'Windows')).returncode == 0:
                hardware['cuda'] = True
            # Also check for nvidia-smi as fallback
            elif subprocess.run(['nvidia-smi'], capture_output=True, shell=(system == 'Windows')).returncode == 0:
                hardware['cuda'] = True
        except (FileNotFoundError, subprocess.SubprocessError):
            # Check for CUDA libraries on Windows
            if system == 'Windows':
                cuda_paths = [
                    os.environ.get('CUDA_PATH'),
                    'C:\\Program Files\\NVIDIA GPU Computing Toolkit\\CUDA',
                    'C:\\Program Files (x86)\\NVIDIA GPU Computing Toolkit\\CUDA'
                ]
                for cuda_path in cuda_paths:
                    if cuda_path and os.path.exists(cuda_path):
                        hardware['cuda'] = True
                        break

        # Check Metal (macOS)
        if system == 'Darwin':
            try:
                # Check for Xcode command line tools
                if subprocess.run(['xcrun', '--show-sdk-path'], capture_output=True).returncode == 0:
                    hardware['metal'] = True
                # Check for Metal framework
                if os.path.exists('/System/Library/Frameworks/Metal.framework'):
                    hardware['metal'] = True
                # macOS also supports Accelerate framework
                if os.path.exists('/System/Library/Frameworks/Accelerate.framework'):
                    hardware['accelerate'] = True
            except (FileNotFoundError, subprocess.SubprocessError):
                pass

        # Check OpenCL
        try:
            if system == 'Windows':
                # Check for OpenCL on Windows
                opencl_paths = [
                    'C:\\Windows\\System32\\OpenCL.dll',
                    'C:\\Windows\\SysWOW64\\OpenCL.dll'
                ]
                if any(os.path.exists(path) for path in opencl_paths):
                    hardware['opencl'] = True
            else:
                if subprocess.run(['clinfo'], capture_output=True).returncode == 0:
                    hardware['opencl'] = True
        except (FileNotFoundError, subprocess.SubprocessError):
            pass

        # Check Vulkan
        try:
            if system == 'Windows':
                # Check for Vulkan on Windows
                vulkan_paths = [
                    'C:\\Windows\\System32\\vulkan-1.dll',
                    'C:\\Windows\\SysWOW64\\vulkan-1.dll'
                ]
                if any(os.path.exists(path) for path in vulkan_paths):
                    hardware['vulkan'] = True
            else:
                if subprocess.run(['vulkaninfo'], capture_output=True).returncode == 0:
                    hardware['vulkan'] = True
        except (FileNotFoundError, subprocess.SubprocessError):
            pass

        # Check ROCm (AMD)
        try:
            if subprocess.run(['rocm-smi'], capture_output=True, shell=(system == 'Windows')).returncode == 0:
                hardware['rocm'] = True
            elif system == 'Linux':
                # Check for ROCm installation
                rocm_paths = ['/opt/rocm', '/usr/lib/x86_64-linux-gnu/librocm-smi64.so']
                if any(os.path.exists(path) for path in rocm_paths):
                    hardware['rocm'] = True
        except (FileNotFoundError, subprocess.SubprocessError):
            pass

        # Check for BLAS libraries
        try:
            import numpy as np  # type: ignore
            # Check if numpy is linked with optimized BLAS
            config = np.__config__.show()
            if any(lib in str(config).lower() for lib in ['openblas', 'mkl', 'atlas', 'blis']):
                hardware['blas'] = True
        except (ImportError, AttributeError):
            # Fallback: check for common BLAS libraries
            if system == 'Linux':
                blas_libs = ['/usr/lib/x86_64-linux-gnu/libopenblas.so', '/usr/lib/x86_64-linux-gnu/libblas.so']
                if any(os.path.exists(lib) for lib in blas_libs):
                    hardware['blas'] = True
            elif system == 'Windows':
                # Check for Intel MKL or OpenBLAS on Windows
                mkl_paths = ['C:\\Program Files (x86)\\Intel\\oneAPI\\mkl']
                if any(os.path.exists(path) for path in mkl_paths):
                    hardware['blas'] = True

        return hardware

    def setup_llama_cpp(self) -> None:
        """Sets up and builds llama.cpp repository with robust error handling."""
        llama_path = self.workspace / "llama.cpp"
        system = platform.system()

        with console.status("[bold green]Setting up llama.cpp...") as status:
            # Clone llama.cpp if not exists
            if not llama_path.exists():
                try:
                    subprocess.run(['git', 'clone', 'https://github.com/ggerganov/llama.cpp'], check=True)
                except subprocess.CalledProcessError as e:
                    raise ConversionError(f"Failed to clone llama.cpp repository: {e}")

            original_cwd = os.getcwd()
            try:
                os.chdir(llama_path)

                # Update to latest version
                try:
                    subprocess.run(['git', 'pull'], capture_output=True, check=False)
                except subprocess.CalledProcessError:
                    console.print("[yellow]Warning: Could not update llama.cpp repository")

                # Clean any existing build directory to avoid cached CMake variables
                build_dir = Path('build')
                if build_dir.exists():
                    console.print("[yellow]Cleaning existing build directory to avoid CMake cache conflicts...")
                    import shutil
                    try:
                        shutil.rmtree(build_dir)
                        console.print("[green]Build directory cleaned successfully")
                    except Exception as e:
                        console.print(f"[yellow]Warning: Could not clean build directory: {e}")

                # Check if we're in a Nix environment
                is_nix = system == "Linux" and os.path.exists("/nix/store")

                if is_nix:
                    console.print("[yellow]Detected Nix environment. Using system Python packages...")
                    # In Nix, we need to use the system Python packages
                    try:
                        # Try to import required packages to check if they're available
                        import numpy  # type: ignore
                        import sentencepiece  # type: ignore
                        import torch  # type: ignore
                        import transformers  # type: ignore
                        console.print("[green]Required Python packages are already installed.")
                    except ImportError:
                        console.print("[red]Missing required Python packages in Nix environment.")
                        console.print("[yellow]Please install them using:")
                        console.print("nix-shell -p python3Packages.torch python3Packages.numpy python3Packages.sentencepiece python3Packages.transformers")
                        raise ConversionError("Missing required Python packages in Nix environment")
                else:
                    # In non-Nix environments, install requirements if they exist
                    if os.path.exists('requirements.txt'):
                        try:
                            pip_cmd = 'pip' if system == 'Windows' else 'pip3'
                            subprocess.run([pip_cmd, 'install', '-r', 'requirements.txt'], check=True)
                        except subprocess.CalledProcessError as e:
                            if "externally-managed-environment" in str(e):
                                console.print("[yellow]Detected externally managed Python environment.")
                                console.print("[yellow]Please install the required packages manually:")
                                console.print("pip install torch numpy sentencepiece transformers")
                                raise ConversionError("Failed to install requirements in externally managed environment")
                            else:
                                console.print(f"[yellow]Warning: Failed to install requirements: {e}")

                # Detect available hardware
                hardware = self.detect_hardware()
                console.print("[bold green]Detected hardware acceleration:")
                for hw, available in hardware.items():
                    console.print(f"  {'✓' if available else '✗'} {hw.upper()}")

                # Clear any environment variables that might cause conflicts
                env_vars_to_clear = [
                    'LLAMA_CUBLAS', 'LLAMA_CLBLAST', 'LLAMA_HIPBLAS',
                    'LLAMA_METAL', 'LLAMA_ACCELERATE', 'LLAMA_OPENBLAS'
                ]
                for var in env_vars_to_clear:
                    if var in os.environ:
                        console.print(f"[yellow]Clearing conflicting environment variable: {var}")
                        del os.environ[var]

                # Configure CMake build with robust options
                cmake_args: List[str] = ['cmake', '-B', 'build']

                # Add basic CMake options
                cmake_args.extend([
                    '-DCMAKE_BUILD_TYPE=Release',
                    '-DLLAMA_BUILD_TESTS=OFF',
                    '-DLLAMA_BUILD_EXAMPLES=ON',
                    '-DLLAMA_BUILD_SERVER=OFF',
                    # Disable optional dependencies that might cause issues
                    '-DLLAMA_CURL=OFF',           # Disable CURL (not needed for GGUF conversion)
                    '-DLLAMA_LLGUIDANCE=OFF'      # Disable LLGuidance (optional feature)
                ])

                # Add hardware acceleration options with latest 2025 llama.cpp GGML flags
                # Use priority order: CUDA > Metal > Vulkan > OpenCL > ROCm > BLAS > Accelerate
                acceleration_enabled = False

                if hardware['cuda']:
                    # Latest 2025 GGML CUDA flags (LLAMA_CUBLAS is deprecated)
                    cmake_args.extend(['-DGGML_CUDA=ON'])
                    console.print("[green]Enabling CUDA acceleration (GGML_CUDA=ON)")
                    acceleration_enabled = True
                elif hardware['metal']:
                    # Latest 2025 GGML Metal flags for macOS
                    cmake_args.extend(['-DGGML_METAL=ON'])
                    console.print("[green]Enabling Metal acceleration (GGML_METAL=ON)")
                    acceleration_enabled = True
                elif hardware['vulkan']:
                    # Latest 2025 GGML Vulkan flags
                    cmake_args.extend(['-DGGML_VULKAN=ON'])
                    console.print("[green]Enabling Vulkan acceleration (GGML_VULKAN=ON)")
                    acceleration_enabled = True
                elif hardware['opencl']:
                    # Latest 2025 GGML OpenCL flags (LLAMA_CLBLAST is deprecated)
                    cmake_args.extend(['-DGGML_OPENCL=ON'])
                    console.print("[green]Enabling OpenCL acceleration (GGML_OPENCL=ON)")
                    acceleration_enabled = True
                elif hardware['rocm']:
                    # Latest 2025 GGML ROCm/HIP flags
                    cmake_args.extend(['-DGGML_HIPBLAS=ON'])
                    console.print("[green]Enabling ROCm acceleration (GGML_HIPBLAS=ON)")
                    acceleration_enabled = True
                elif hardware['blas']:
                    # Latest 2025 GGML BLAS flags with vendor detection
                    cmake_args.extend(['-DGGML_BLAS=ON'])
                    # Try to detect BLAS vendor for optimal performance
                    if system == 'Darwin':
                        cmake_args.extend(['-DGGML_BLAS_VENDOR=Accelerate'])
                    elif 'mkl' in str(hardware).lower():
                        cmake_args.extend(['-DGGML_BLAS_VENDOR=Intel10_64lp'])
                    else:
                        cmake_args.extend(['-DGGML_BLAS_VENDOR=OpenBLAS'])
                    console.print("[green]Enabling BLAS acceleration (GGML_BLAS=ON)")
                    acceleration_enabled = True
                elif hardware['accelerate']:
                    # Latest 2025 GGML Accelerate framework flags for macOS
                    cmake_args.extend(['-DGGML_ACCELERATE=ON'])
                    console.print("[green]Enabling Accelerate framework (GGML_ACCELERATE=ON)")
                    acceleration_enabled = True

                if not acceleration_enabled:
                    console.print("[yellow]No hardware acceleration available, using CPU only")

                # Platform-specific optimizations
                if system == 'Windows':
                    # Use Visual Studio generator on Windows if available
                    try:
                        vs_result = subprocess.run(['where', 'msbuild'], capture_output=True)
                        if vs_result.returncode == 0:
                            cmake_args.extend(['-G', 'Visual Studio 17 2022'])
                        else:
                            cmake_args.extend(['-G', 'MinGW Makefiles'])
                    except Exception:
                        cmake_args.extend(['-G', 'MinGW Makefiles'])
                else:
                    # Use Ninja if available on Unix systems
                    try:
                        ninja_cmd = 'ninja' if system != 'Windows' else 'ninja.exe'
                        if subprocess.run(['which', ninja_cmd], capture_output=True).returncode == 0:
                            cmake_args.extend(['-G', 'Ninja'])
                    except Exception:
                        pass  # Fall back to default generator

                # Configure the build with error handling and multiple fallback strategies
                status.update("[bold green]Configuring CMake build...")
                config_success = False

                # Try main configuration
                try:
                    console.print(f"[cyan]CMake command: {' '.join(cmake_args)}")
                    result = subprocess.run(cmake_args, capture_output=True, text=True)
                    if result.returncode == 0:
                        config_success = True
                        console.print("[green]CMake configuration successful!")
                    else:
                        console.print(f"[red]CMake configuration failed: {result.stderr}")
                except subprocess.CalledProcessError as e:
                    console.print(f"[red]CMake execution failed: {e}")

                # Try fallback without hardware acceleration if main config failed
                if not config_success:
                    console.print("[yellow]Attempting fallback configuration without hardware acceleration...")
                    console.print("[cyan]Using CPU-only build configuration...")
                    fallback_args = [
                        'cmake', '-B', 'build',
                        '-DCMAKE_BUILD_TYPE=Release',
                        '-DLLAMA_BUILD_TESTS=OFF',
                        '-DLLAMA_BUILD_EXAMPLES=ON',
                        '-DLLAMA_BUILD_SERVER=OFF',
                        # Disable optional dependencies that might cause issues
                        '-DLLAMA_CURL=OFF',           # Disable CURL (not needed for GGUF conversion)
                        '-DLLAMA_LLGUIDANCE=OFF',     # Disable LLGuidance (optional feature)
                        # Enable CPU optimizations
                        '-DGGML_NATIVE=OFF',  # Disable native optimizations for compatibility
                        '-DGGML_AVX=ON',      # Enable AVX if available
                        '-DGGML_AVX2=ON',     # Enable AVX2 if available
                        '-DGGML_FMA=ON'       # Enable FMA if available
                    ]
                    try:
                        console.print(f"[cyan]Fallback CMake command: {' '.join(fallback_args)}")
                        result = subprocess.run(fallback_args, capture_output=True, text=True)
                        if result.returncode == 0:
                            config_success = True
                            console.print("[green]Fallback CMake configuration successful!")
                        else:
                            console.print(f"[red]Fallback CMake configuration failed: {result.stderr}")
                    except subprocess.CalledProcessError as e:
                        console.print(f"[red]Fallback CMake execution failed: {e}")

                # Try minimal configuration as last resort
                if not config_success:
                    console.print("[yellow]Attempting minimal configuration...")
                    minimal_args = [
                        'cmake', '-B', 'build',
                        '-DCMAKE_BUILD_TYPE=Release',
                        # Disable optional dependencies that might cause issues
                        '-DLLAMA_CURL=OFF',           # Disable CURL (not needed for GGUF conversion)
                        '-DLLAMA_LLGUIDANCE=OFF',     # Disable LLGuidance (optional feature)
                        '-DLLAMA_BUILD_SERVER=OFF',   # Disable server (not needed for conversion)
                        '-DLLAMA_BUILD_TESTS=OFF'     # Disable tests (not needed for conversion)
                    ]
                    try:
                        console.print(f"[cyan]Minimal CMake command: {' '.join(minimal_args)}")
                        result = subprocess.run(minimal_args, capture_output=True, text=True)
                        if result.returncode == 0:
                            config_success = True
                            console.print("[green]Minimal CMake configuration successful!")
                        else:
                            console.print(f"[red]Minimal CMake configuration failed: {result.stderr}")
                            raise ConversionError(f"All CMake configuration attempts failed. Last error: {result.stderr}")
                    except subprocess.CalledProcessError as e:
                        raise ConversionError(f"All CMake configuration attempts failed: {e}")

                if not config_success:
                    raise ConversionError("CMake configuration failed with all attempted strategies")

                # Build the project
                status.update("[bold green]Building llama.cpp...")
                build_cmd = ['cmake', '--build', 'build', '--config', 'Release']

                # Add parallel build option
                cpu_count = os.cpu_count() or 1
                if system == 'Windows':
                    build_cmd.extend(['--parallel', str(cpu_count)])
                else:
                    build_cmd.extend(['-j', str(cpu_count)])

                try:
                    result = subprocess.run(build_cmd, capture_output=True, text=True)
                    if result.returncode != 0:
                        console.print(f"[red]Build failed: {result.stderr}")
                        # Try single-threaded build as fallback
                        console.print("[yellow]Attempting single-threaded build...")
                        fallback_build = ['cmake', '--build', 'build', '--config', 'Release']
                        result = subprocess.run(fallback_build, capture_output=True, text=True)
                        if result.returncode != 0:
                            raise ConversionError(f"Build failed: {result.stderr}")
                except subprocess.CalledProcessError as e:
                    raise ConversionError(f"Build failed: {e}")

                console.print("[green]llama.cpp built successfully!")

            finally:
                os.chdir(original_cwd)

    def display_config(self) -> None:
        """Displays the current configuration in a formatted table."""
        table = Table(title="Configuration", show_header=True, header_style="bold magenta")
        table.add_column("Setting", style="cyan")
        table.add_column("Value", style="green")

        table.add_row("Model ID", self.model_id)
        table.add_row("Model Name", self.model_name)
        table.add_row("Username", self.username or "Not provided")
        table.add_row("Token", "****" if self.token else "Not provided")
        table.add_row("Quantization Methods", "\n".join(
            f"{method} ({self.VALID_METHODS[method]})"
            for method in self.quantization_methods
        ))

        console.print(Panel(table))

    def get_binary_path(self, binary_name: str) -> str:
        """Get the correct path to llama.cpp binaries based on platform."""
        system = platform.system()

        # Possible binary locations
        possible_paths = [
            f"./llama.cpp/build/bin/{binary_name}",  # Standard build location
            f"./llama.cpp/build/{binary_name}",      # Alternative build location
            f"./llama.cpp/{binary_name}",            # Root directory
            f"./llama.cpp/build/Release/{binary_name}",  # Windows Release build
            f"./llama.cpp/build/Debug/{binary_name}",    # Windows Debug build
        ]

        # Add .exe extension on Windows
        if system == 'Windows':
            possible_paths = [path + '.exe' for path in possible_paths]

        # Find the first existing binary
        for path in possible_paths:
            if os.path.isfile(path):
                return path

        # If not found, return the most likely path and let it fail with a clear error
        default_path = f"./llama.cpp/build/bin/{binary_name}"
        if system == 'Windows':
            default_path += '.exe'
        return default_path

    def generate_importance_matrix(self, model_path: str, train_data_path: str, output_path: str) -> None:
        """Generates importance matrix for quantization with improved error handling."""
        imatrix_binary = self.get_binary_path("llama-imatrix")

        imatrix_command: List[str] = [
            imatrix_binary,
            "-m", model_path,
            "-f", train_data_path,
            "-ngl", "99",
            "--output-frequency", "10",
            "-o", output_path,
        ]

        if not os.path.isfile(model_path):
            raise ConversionError(f"Model file not found: {model_path}")

        if not os.path.isfile(train_data_path):
            raise ConversionError(f"Training data file not found: {train_data_path}")

        if not os.path.isfile(imatrix_binary):
            raise ConversionError(f"llama-imatrix binary not found at: {imatrix_binary}")

        console.print("[bold green]Generating importance matrix...")
        console.print(f"[cyan]Command: {' '.join(imatrix_command)}")

        try:
            process = subprocess.Popen(
                imatrix_command,
                shell=False,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True
            )

            try:
                stdout, stderr = process.communicate(timeout=300)  # 5 minute timeout
                if process.returncode != 0:
                    raise ConversionError(f"Failed to generate importance matrix: {stderr}")
            except subprocess.TimeoutExpired:
                console.print("[yellow]Imatrix computation timed out. Sending SIGINT...")
                process.send_signal(signal.SIGINT)
                try:
                    stdout, stderr = process.communicate(timeout=10)
                except subprocess.TimeoutExpired:
                    console.print("[red]Imatrix process still running. Force terminating...")
                    process.kill()
                    stdout, stderr = process.communicate()
                raise ConversionError(f"Imatrix generation timed out: {stderr}")
        except FileNotFoundError:
            raise ConversionError(f"Could not execute llama-imatrix binary: {imatrix_binary}")

        console.print("[green]Importance matrix generation completed.")

    def split_model(self, model_path: str, outdir: str) -> List[str]:
        """Splits the model into smaller chunks with improved error handling."""
        split_binary = self.get_binary_path("llama-gguf-split")

        split_cmd: List[str] = [
            split_binary,
            "--split",
        ]

        if self.split_max_size:
            split_cmd.extend(["--split-max-size", self.split_max_size])
        else:
            split_cmd.extend(["--split-max-tensors", str(self.split_max_tensors)])

        model_path_prefix = '.'.join(model_path.split('.')[:-1])
        split_cmd.extend([model_path, model_path_prefix])

        if not os.path.isfile(model_path):
            raise ConversionError(f"Model file not found: {model_path}")

        if not os.path.isfile(split_binary):
            raise ConversionError(f"llama-gguf-split binary not found at: {split_binary}")

        console.print(f"[bold green]Splitting model with command: {' '.join(split_cmd)}")

        try:
            result = subprocess.run(split_cmd, shell=False, capture_output=True, text=True)

            if result.returncode != 0:
                raise ConversionError(f"Error splitting model: {result.stderr}")
        except FileNotFoundError:
            raise ConversionError(f"Could not execute llama-gguf-split binary: {split_binary}")

        console.print("[green]Model split successfully!")

        # Get list of split files
        model_file_prefix = os.path.basename(model_path_prefix)
        try:
            split_files = [f for f in os.listdir(outdir)
                          if f.startswith(model_file_prefix) and f.endswith(".gguf")]
        except OSError as e:
            raise ConversionError(f"Error reading output directory: {e}")

        if not split_files:
            raise ConversionError(f"No split files found in {outdir} with prefix {model_file_prefix}")

        console.print(f"[green]Found {len(split_files)} split files: {', '.join(split_files)}")
        return split_files

    def upload_split_files(self, split_files: List[str], outdir: str, repo_id: str) -> None:
        """Uploads split model files to Hugging Face."""
        api = _ensure_huggingface_hub()(token=self.token)

        for file in split_files:
            file_path = os.path.join(outdir, file)
            console.print(f"[bold green]Uploading file: {file}")
            try:
                api.upload_file(
                    path_or_fileobj=file_path,
                    path_in_repo=file,
                    repo_id=repo_id,
                )
                console.print(f"[green]✓ Successfully uploaded: {file}")
            except Exception as e:
                console.print(f"[red]✗ Failed to upload {file}: {e}")
                raise ConversionError(f"Error uploading file {file}: {e}")

    def generate_readme(self, quantized_files: List[str]) -> str:
        """Generate a README.md file for the Hugging Face Hub."""
        readme = f"""# {self.model_name} GGUF

This repository contains GGUF quantized versions of [{self.model_id}](https://huggingface.co/{self.model_id}).

## About

This model was converted using [Webscout](https://github.com/Webscout/webscout).

## Quantization Methods

The following quantization methods were used:

"""
        # Add quantization method descriptions
        for method in self.quantization_methods:
            if self.use_imatrix:
                readme += f"- `{method}`: {self.VALID_IMATRIX_METHODS[method]}\n"
            else:
                readme += f"- `{method}`: {self.VALID_METHODS[method]}\n"

        readme += """
## Available Files

The following quantized files are available:

"""
        # Add file information
        for file in quantized_files:
            readme += f"- `{file}`\n"

        if self.use_imatrix:
            readme += """
## Importance Matrix

This model was quantized using importance matrix quantization. The `imatrix.dat` file contains the importance matrix used for quantization.

"""

        readme += """
## Usage

These GGUF files can be used with [llama.cpp](https://github.com/ggerganov/llama.cpp) and compatible tools.

Example usage:
```bash
./main -m model.gguf -n 1024 --repeat_penalty 1.0 --color -i -r "User:" -f prompts/chat-with-bob.txt
```

## Conversion Process

This model was converted using the following command:
```bash
python -m webscout.Extra.gguf convert \\
    -m "{self.model_id}" \\
    -q "{','.join(self.quantization_methods)}" \\
    {f'-i' if self.use_imatrix else ''} \\
    {f'--train-data "{self.train_data_file}"' if self.train_data_file else ''} \\
    {f'-s' if self.use_split else ''} \\
    {f'--split-max-tensors {self.split_max_tensors}' if self.use_split else ''} \\
    {f'--split-max-size {self.split_max_size}' if self.split_max_size else ''}
```

## License

This repository is licensed under the same terms as the original model.
"""
        return readme

    def create_repository(self, repo_id: str) -> None:
        """Create a new repository on Hugging Face Hub if it doesn't exist."""
        api = _ensure_huggingface_hub()(token=self.token)
        try:
            # Check if repository already exists
            try:
                api.repo_info(repo_id=repo_id)
                console.print(f"[green]✓ Repository {repo_id} already exists")
                return
            except Exception:
                # Repository doesn't exist, create it
                pass

            console.print(f"[bold green]Creating new repository: {repo_id}")
            api.create_repo(
                repo_id=repo_id,
                exist_ok=True,
                private=False,
                repo_type="model"
            )
            console.print(f"[green]✓ Successfully created repository: {repo_id}")
            console.print(f"[cyan]Repository URL: https://huggingface.co/{repo_id}")
        except Exception as e:
            console.print(f"[red]✗ Failed to create repository: {e}")
            raise ConversionError(f"Error creating repository {repo_id}: {e}")

    def upload_readme(self, readme_content: str, repo_id: str) -> None:
        """Upload README.md to Hugging Face Hub."""
        api = _ensure_huggingface_hub()(token=self.token)
        console.print("[bold green]Uploading README.md with model documentation")
        try:
            api.upload_file(
                path_or_fileobj=readme_content.encode(),
                path_in_repo="README.md",
                repo_id=repo_id,
            )
            console.print("[green]✓ Successfully uploaded: README.md")
        except Exception as e:
            console.print(f"[red]✗ Failed to upload README.md: {e}")
            raise ConversionError(f"Error uploading README.md: {e}")

    def convert(self) -> None:
        """Performs the model conversion process."""
        try:
            # Display banner and configuration
            console.print(f"[bold green]{figlet_format('GGUF Converter')}")
            self.display_config()

            # Validate inputs
            self.validate_inputs()

            # Check dependencies
            deps = self.check_dependencies()
            missing = [name for name, installed in deps.items() if not installed and name != 'ninja']
            if missing:
                raise ConversionError(f"Missing required dependencies: {', '.join(missing)}")

            # Setup llama.cpp
            self.setup_llama_cpp()

            # Determine if we need temporary directories (only for uploads)
            needs_temp = bool(self.username and self.token)

            if needs_temp:
                # Use temporary directories for upload case
                with tempfile.TemporaryDirectory() as outdir:
                    with tempfile.TemporaryDirectory() as tmpdir:
                        self._convert_with_dirs(tmpdir, outdir)
            else:
                # Use current directory for local output
                outdir = os.getcwd()
                tmpdir = os.path.join(outdir, "temp_download")
                os.makedirs(tmpdir, exist_ok=True)
                try:
                    self._convert_with_dirs(tmpdir, outdir)
                finally:
                    # Clean up temporary download directory
                    import shutil
                    shutil.rmtree(tmpdir, ignore_errors=True)

            # Display success message
            console.print(Panel.fit(
                "[bold green]✓[/] Conversion completed successfully!\n\n"
                f"[cyan]Output files can be found in: {self.workspace / self.model_name}[/]",
                title="Success",
                border_style="green"
            ))

        except Exception as e:
            console.print(Panel.fit(
                f"[bold red]✗[/] {str(e)}",
                title="Error",
                border_style="red"
            ))
            raise

    def _convert_with_dirs(self, tmpdir: str, outdir: str) -> None:
        """Helper method to perform conversion with given directories."""
        # Use outtype for base filename (e.g., model.f16.gguf, model.bf16.gguf)
        outtype_suffix = self.outtype if self.outtype != "auto" else "f16"
        base_gguf = str(Path(outdir)/f"{self.model_name}.{outtype_suffix}.gguf")

        # Download model (or use remote mode)
        local_dir = Path(tmpdir)/self.model_name
        if self.remote:
            console.print("[bold green]Using remote mode - downloading only config and tokenizer...")
            api = _ensure_huggingface_hub()(token=self.token)
            api.snapshot_download(
                repo_id=self.model_id,
                local_dir=local_dir,
                allow_patterns=["LICENSE", "*.json", "*.md", "*.txt", "tokenizer.model"]
            )
        else:
            console.print("[bold green]Downloading model...")
            api = _ensure_huggingface_hub()(token=self.token)
            api.snapshot_download(
                repo_id=self.model_id,
                local_dir=local_dir,
            )

        # Convert to GGUF with specified outtype
        console.print(f"[bold green]Converting to {self.outtype}...")

        # Find the conversion script
        conversion_scripts = [
            "llama.cpp/convert_hf_to_gguf.py",
            "llama.cpp/convert-hf-to-gguf.py",
            "llama.cpp/convert.py"
        ]

        conversion_script = None
        for script in conversion_scripts:
            if os.path.isfile(script):
                conversion_script = script
                break

        if not conversion_script:
            raise ConversionError("Could not find HuggingFace to GGUF conversion script")

        # Use the appropriate Python executable
        python_cmd = "python" if platform.system() == "Windows" else "python3"

        # Build conversion command with new llama.cpp options
        convert_cmd = [
            python_cmd, conversion_script,
            str(local_dir),
            "--outtype", self.outtype,
            "--outfile", base_gguf
        ]

        # Add optional flags based on new llama.cpp features
        if self.vocab_only:
            convert_cmd.append("--vocab-only")
        if self.no_lazy:
            convert_cmd.append("--no-lazy")
        if self.dry_run:
            convert_cmd.append("--dry-run")
        if self.remote:
            convert_cmd.extend(["--remote"])
        if self.model_name:
            convert_cmd.extend(["--model-name", self.model_name])
        if self.use_split:
            if self.split_max_tensors > 0:
                convert_cmd.extend(["--split-max-tensors", str(self.split_max_tensors)])
            if self.split_max_size:
                convert_cmd.extend(["--split-max-size", self.split_max_size])
        if self.small_first_shard:
            convert_cmd.append("--no-tensor-first-split")

        console.print(f"[cyan]Conversion command: {' '.join(convert_cmd)}")

        try:
            result = subprocess.run(convert_cmd, capture_output=True, text=True)

            if result.returncode != 0:
                raise ConversionError(f"Error converting to {self.outtype}: {result.stderr}")
        except FileNotFoundError as e:
            raise ConversionError(f"Could not execute conversion script: {e}")

        if not os.path.isfile(base_gguf):
            raise ConversionError(f"Conversion completed but output file not found: {base_gguf}")

        console.print(f"[green]Model converted to {self.outtype} successfully!")

        # If base_only is True, we're done after base conversion (no quantization needed)
        if self.base_only:
            gguf_filename = f"{self.model_name}.{outtype_suffix}.gguf"
            quantized_files = [gguf_filename]
            if self.username and self.token:
                repo_id = f"{self.username}/{self.model_name}-GGUF"

                # Step 1: Create repository
                self.create_repository(repo_id)

                # Step 2: Upload README first
                readme_content = self.generate_readme(quantized_files)
                self.upload_readme(readme_content, repo_id)

                # Step 3: Upload model GGUF file
                console.print(f"[bold green]Uploading model file: {gguf_filename}")
                try:
                    api.upload_file(
                        path_or_fileobj=base_gguf,
                        path_in_repo=gguf_filename,
                        repo_id=repo_id
                    )
                    console.print(f"[green]✓ Successfully uploaded: {gguf_filename}")
                except Exception as e:
                    console.print(f"[red]✗ Failed to upload {gguf_filename}: {e}")
                    raise ConversionError(f"Error uploading model file: {e}")
            return

        # Generate importance matrix if needed
        imatrix_path: Optional[str] = None
        if self.use_imatrix:
            train_data_path = self.train_data_file if self.train_data_file else "llama.cpp/groups_merged.txt"
            imatrix_path = str(Path(outdir)/"imatrix.dat")
            self.generate_importance_matrix(base_gguf, train_data_path, imatrix_path)

        # Quantize model
        console.print("[bold green]Quantizing model...")
        quantized_files: List[str] = []
        quantize_binary = self.get_binary_path("llama-quantize")
        quantized_path = ""

        if not os.path.isfile(quantize_binary):
            raise ConversionError(f"llama-quantize binary not found at: {quantize_binary}")

        for method in self.quantization_methods:
            quantized_name = f"{self.model_name.lower()}-{method.lower()}"
            if self.use_imatrix:
                quantized_name += "-imat"
            quantized_path = str(Path(outdir)/f"{quantized_name}.gguf")

            console.print(f"[cyan]Quantizing with method: {method}")

            if self.use_imatrix and imatrix_path:
                quantize_cmd: List[str] = [
                    quantize_binary,
                    "--imatrix", str(imatrix_path),
                    base_gguf, quantized_path, method
                ]
            else:
                quantize_cmd = [
                    quantize_binary,
                    base_gguf, quantized_path, method
                ]

            console.print(f"[cyan]Quantization command: {' '.join(quantize_cmd)}")

            try:
                result = subprocess.run(quantize_cmd, capture_output=True, text=True)
                if result.returncode != 0:
                    raise ConversionError(f"Error quantizing with {method}: {result.stderr}")
            except FileNotFoundError:
                raise ConversionError(f"Could not execute llama-quantize binary: {quantize_binary}")

            if not os.path.isfile(quantized_path):
                raise ConversionError(f"Quantization completed but output file not found: {quantized_path}")

            quantized_files.append(f"{quantized_name}.gguf")
            console.print(f"[green]Successfully quantized with {method}: {quantized_name}.gguf")

        # Upload to Hugging Face if credentials provided
        if self.username and self.token:
            repo_id = f"{self.username}/{self.model_name}-GGUF"

            # Step 1: Create repository
            console.print(f"[bold blue]Step 1: Creating repository {repo_id}")
            self.create_repository(repo_id)

            # Step 2: Generate and upload README first
            console.print("[bold blue]Step 2: Uploading README.md")
            readme_content = self.generate_readme(quantized_files)
            self.upload_readme(readme_content, repo_id)

            # Step 3: Upload model GGUF files
            console.print("[bold blue]Step 3: Upload model files")
            if self.use_split:
                split_files = self.split_model(quantized_path, outdir)
                self.upload_split_files(split_files, outdir, repo_id)
            else:
                # Upload single quantized file
                file_name = f"{self.model_name.lower()}-{self.quantization_methods[0].lower()}.gguf"
                console.print(f"[bold green]Uploading quantized model: {file_name}")
                try:
                    api.upload_file(
                        path_or_fileobj=quantized_path,
                        path_in_repo=file_name,
                        repo_id=repo_id
                    )
                    console.print(f"[green]✓ Successfully uploaded: {file_name}")
                except Exception as e:
                    console.print(f"[red]✗ Failed to upload {file_name}: {e}")
                    raise ConversionError(f"Error uploading quantized model: {e}")

            # Step 4: Upload imatrix if generated (optional)
            if imatrix_path:
                console.print("[bold blue]Step 4: Uploading importance matrix")
                console.print("[bold green]Uploading importance matrix: imatrix.dat")
                try:
                    api.upload_file(
                        path_or_fileobj=imatrix_path,
                        path_in_repo="imatrix.dat",
                        repo_id=repo_id
                    )
                    console.print("[green]✓ Successfully uploaded: imatrix.dat")
                except Exception as e:
                    console.print(f"[yellow]Warning: Failed to upload imatrix.dat: {e}")

            # Final success message
            console.print(f"[bold green]🎉 All files uploaded successfully to {repo_id}!")
            console.print(f"[cyan]Repository URL: https://huggingface.co/{repo_id}")

# Initialize CLI with HAI vibes
app = CLI(
    name="gguf",
    help="Convert HuggingFace models to GGUF format with style! 🔥",
    version="2.0.0"
)

@app.command(name="convert")
@option("-m", "--model-id", help="The HuggingFace model ID (e.g., 'OEvortex/HelpingAI-Lite-1.5T')", required=True)
@option("-u", "--username", help="Your HuggingFace username for uploads", default=None)
@option("-t", "--token", help="Your HuggingFace API token for uploads", default=None)
@option("-q", "--quantization", help="Comma-separated quantization methods", default="q4_k_m")
@option("-o", "--outtype", help="Output type: f32, f16, bf16, q8_0, tq1_0, tq2_0, auto", default="f16")
@option("-i", "--use-imatrix", help="Use importance matrix for quantization", is_flag=True)
@option("--train-data", help="Training data file for imatrix quantization", default=None)
@option("-s", "--split-model", help="Split the model into smaller chunks", is_flag=True)
@option("--split-max-tensors", help="Maximum number of tensors per file when splitting", default=256)
@option("--split-max-size", help="Maximum file size when splitting (e.g., '256M', '5G')", default=None)
@option("--vocab-only", help="Only extract vocabulary (no model weights)", is_flag=True)
@option("--remote", help="(Experimental) Read tensors remotely without full download", is_flag=True)
@option("--dry-run", help="Only print split plan without writing files", is_flag=True)
@option("--no-lazy", help="Disable lazy evaluation (use more RAM)", is_flag=True)
@option("--model-name", help="Custom model name override", default=None)
@option("--small-first-shard", help="Do not add tensors to the first split", is_flag=True)
def convert_command(
    model_id: str,
    username: Optional[str] = None,
    token: Optional[str] = None,
    quantization: str = "q4_k_m",
    outtype: str = "f16",
    use_imatrix: bool = False,
    train_data: Optional[str] = None,
    split_model: bool = False,
    split_max_tensors: int = 256,
    split_max_size: Optional[str] = None,
    vocab_only: bool = False,
    remote: bool = False,
    dry_run: bool = False,
    no_lazy: bool = False,
    model_name: Optional[str] = None,
    small_first_shard: bool = False
) -> None:
    """
    Convert and quantize HuggingFace models to GGUF format! 🚀

    Args:
        model_id (str): Your model's HF ID (like 'OEvortex/HelpingAI-Lite-1.5T') 🎯
        username (str, optional): Your HF username for uploads 👤
        token (str, optional): Your HF API token 🔑
        quantization (str): Quantization methods (default: q4_k_m) 🎮
        outtype (str): Output type: f32, f16, bf16, q8_0, tq1_0, tq2_0, auto 📊
        use_imatrix (bool): Use importance matrix for quantization 🔍
        train_data (str, optional): Training data file for imatrix quantization 📚
        split_model (bool): Split the model into smaller chunks 🔪
        split_max_tensors (int): Max tensors per file when splitting (default: 256) 📊
        split_max_size (str, optional): Max file size when splitting (e.g., '256M', '5G') 📏
        vocab_only (bool): Only extract vocabulary (no model weights) 📖
        remote (bool): (Experimental) Read tensors remotely without full download 🌐
        dry_run (bool): Only print split plan without writing files 📝
        no_lazy (bool): Disable lazy evaluation (use more RAM) 🧠
        model_name (str, optional): Custom model name override 🏷️
        small_first_shard (bool): Do not add tensors to the first split 📦

    Example:
        >>> python -m webscout.Extra.gguf convert \\
        ...     -m "OEvortex/HelpingAI-Lite-1.5T" \\
        ...     -q "q4_k_m,q5_k_m" \\
        ...     -o "f16"
    """
    try:
        converter = ModelConverter(
            model_id=model_id,
            username=username,
            token=token,
            quantization_methods=quantization,
            use_imatrix=use_imatrix,
            train_data_file=train_data,
            split_model=split_model,
            split_max_tensors=split_max_tensors,
            split_max_size=split_max_size,
            outtype=outtype,
            vocab_only=vocab_only,
            remote=remote,
            dry_run=dry_run,
            no_lazy=no_lazy,
            model_name=model_name,
            small_first_shard=small_first_shard
        )
        converter.convert()
    except (ConversionError, ValueError) as e:
        console.print(f"[red]Error: {str(e)}")
        sys.exit(1)
    except Exception as e:
        console.print(f"[red]Unexpected error: {str(e)}")
        sys.exit(1)

def main() -> None:
    """Fire up the GGUF converter! 🚀"""
    app.run()

if __name__ == "__main__":
    main()



"""
TypliAI OpenAI-compatible provider.
"""

import random
import string
import time
import uuid
from typing import Any, Dict, Generator, List, Optional, Union, cast

from curl_cffi.requests import Session

from webscout.AIutel import sanitize_stream
from webscout.Provider.Openai_comp.base import (
    BaseChat,
    BaseCompletions,
    OpenAICompatibleProvider,
    SimpleModelList,
)
from webscout.Provider.Openai_comp.utils import (
    ChatCompletion,
    ChatCompletionChunk,
    ChatCompletionMessage,
    Choice,
    ChoiceDelta,
    CompletionUsage,
    format_prompt,
)

try:
    from ...litagent import LitAgent
except ImportError:
    LitAgent = None  # type: ignore


def generate_random_id(length=16):
    """Generates a random alphanumeric string."""
    characters = string.ascii_letters + string.digits
    return "".join(random.choice(characters) for i in range(length))


class Completions(BaseCompletions):
    def __init__(self, client: "TypliAI"):
        self._client = client

    def create(
        self,
        *,
        model: str = "openai/gpt-4.1-mini",
        messages: List[Dict[str, str]],
        stream: bool = False,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        max_tokens: Optional[int] = None,
        timeout: Optional[int] = None,
        proxies: Optional[Dict[str, str]] = None,
        **kwargs: Any,
    ) -> Union[ChatCompletion, Generator[ChatCompletionChunk, None, None]]:
        """
        Creates a model response for the given chat conversation.
        """
        request_id = f"chatcmpl-{uuid.uuid4()}"
        created_time = int(time.time())

        # Use format_prompt from utils
        prompt = format_prompt(messages)

        if stream:
            return self._create_stream(request_id, created_time, model, prompt, timeout, proxies)
        else:
            return self._create_non_stream(
                request_id, created_time, model, prompt, timeout, proxies
            )

    def _create_stream(
        self,
        request_id: str,
        created_time: int,
        model: str,
        prompt: str,
        timeout: Optional[int] = None,
        proxies: Optional[Dict[str, str]] = None,
    ) -> Generator[ChatCompletionChunk, None, None]:
        """Implementation for streaming chat completions."""
        try:
            payload = {
                "slug": "free-no-sign-up-chatgpt",
                "modelId": model,
                "id": generate_random_id(),
                "messages": [
                    {
                        "id": generate_random_id(),
                        "role": "user",
                        "parts": [{"type": "text", "text": prompt}],
                    }
                ],
                "trigger": "submit-message",
            }

            response = self._client.session.post(
                self._client.api_endpoint,
                headers=self._client.headers,
                json=payload,
                stream=True,
                timeout=timeout or self._client.timeout,
                impersonate="chrome120",
                proxies=proxies or self._client.proxies,
            )

            if not response.ok:
                raise IOError(f"TypliAI request failed: {response.status_code} - {response.text}")

            full_response = ""
            completion_tokens = 0

            # Use chunks from iter_content
            data_generator = response.iter_content(chunk_size=None)

            processed_stream = sanitize_stream(
                data=data_generator,
                intro_value="data: ",
                to_json=True,
                content_extractor=lambda x: x.get("delta")
                if isinstance(x, dict) and x.get("type") == "text-delta"
                else None,
                skip_markers=["[DONE]"],
            )

            for content in processed_stream:
                if content and isinstance(content, str):
                    full_response += content
                    completion_tokens += 1

                    delta = ChoiceDelta(
                        content=content, role="assistant" if completion_tokens == 1 else None
                    )
                    choice = Choice(index=0, delta=delta, finish_reason=None)
                    chunk = ChatCompletionChunk(
                        id=request_id, choices=[choice], created=created_time, model=model
                    )
                    yield chunk

            # Final chunk
            choice = Choice(index=0, delta=ChoiceDelta(content=None), finish_reason="stop")
            yield ChatCompletionChunk(
                id=request_id, choices=[choice], created=created_time, model=model
            )

        except Exception as e:
            raise IOError(f"TypliAI stream request failed: {e}") from e

    def _create_non_stream(
        self,
        request_id: str,
        created_time: int,
        model: str,
        prompt: str,
        timeout: Optional[int] = None,
        proxies: Optional[Dict[str, str]] = None,
    ) -> ChatCompletion:
        """Implementation for non-streaming chat completions."""
        full_content = ""
        for chunk in self._create_stream(request_id, created_time, model, prompt, timeout, proxies):
            if chunk.choices[0].delta and chunk.choices[0].delta.content:
                full_content += chunk.choices[0].delta.content

        message = ChatCompletionMessage(role="assistant", content=full_content)
        choice = Choice(index=0, message=message, finish_reason="stop")
        usage = CompletionUsage(
            prompt_tokens=len(prompt.split()),
            completion_tokens=len(full_content.split()),
            total_tokens=len(prompt.split()) + len(full_content.split()),
        )

        return ChatCompletion(
            id=request_id, choices=[choice], created=created_time, model=model, usage=usage
        )


class Chat(BaseChat):
    def __init__(self, client: "TypliAI"):
        self.completions = Completions(client)


class TypliAI(OpenAICompatibleProvider):
    """
    OpenAI-compatible client for TypliAI.
    """

    required_auth = False

    AVAILABLE_MODELS = [
        "openai/gpt-4.1-mini",
        "openai/gpt-4.1",
        "openai/gpt-5-mini",
        "openai/gpt-5.2",
        "openai/gpt-5.2-pro",
        "google/gemini-2.5-flash",
        "anthropic/claude-haiku-4-5",
        "xai/grok-4-fast-reasoning",
        "xai/grok-4-fast",
    ]

    def __init__(self, timeout: int = 60, proxies: Optional[dict] = None, browser: str = "chrome"):
        self.timeout = timeout
        self.proxies = proxies or {}
        self.api_endpoint = "https://typli.ai/api/generators/chat"

        self.session = Session()
        self.agent = LitAgent() if LitAgent else None
        user_agent = (
            self.agent.random()
            if self.agent
            else "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36"
        )

        self.headers = {
            "accept": "/",
            "accept-language": "en-US,en;q=0.9,en-IN;q=0.8",
            "content-type": "application/json",
            "dnt": "1",
            "origin": "https://typli.ai",
            "priority": "u=1, i",
            "referer": "https://typli.ai/free-no-sign-up-chatgpt",
            "sec-ch-ua": '"Not(A:Brand";v="8", "Chromium";v="120", "Google Chrome";v="120"',
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": '"Windows"',
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "sec-gpc": "1",
            "user-agent": user_agent,
        }

        self.session.headers.update(self.headers)
        if proxies:
            if proxies:
                self.session.proxies.update(cast(Any, proxies))

        self.chat = Chat(self)

    @property
    def models(self) -> SimpleModelList:
        return SimpleModelList(type(self).AVAILABLE_MODELS)


if __name__ == "__main__":
    client = TypliAI()
    print(f"Available models: {client.models.list()}")

    # Test non-streaming
    print("\n=== Testing Non-Streaming ===")
    response = client.chat.completions.create(
        model="openai/gpt-4.1-mini",
        messages=[{"role": "user", "content": "Hello! How are you?"}],
        stream=False,
    )
    if isinstance(response, ChatCompletion):
        if response.choices[0].message and response.choices[0].message.content:
            print(f"Response: {response.choices[0].message.content}")
    else:
        print(f"Response: {response}")

    # Test streaming
    print("\n=== Testing Streaming ===")
    stream_resp = client.chat.completions.create(
        model="openai/gpt-4.1-mini",
        messages=[{"role": "user", "content": "Tell me a joke"}],
        stream=True,
    )
    if hasattr(stream_resp, "__iter__") and not isinstance(
        stream_resp, (str, bytes, ChatCompletion)
    ):
        for chunk in cast(Generator[ChatCompletionChunk, None, None], stream_resp):
            if chunk.choices[0].delta and chunk.choices[0].delta.content:
                print(chunk.choices[0].delta.content, end="", flush=True)
    else:
        print(stream_resp)
    print()



"""YouTube search suggestions and autocomplete."""
import json
from typing import List, Optional
from urllib.parse import quote
from urllib.request import Request, urlopen

try:
    from webscout.litagent.agent import LitAgent
    _USER_AGENT = LitAgent().random()
except ImportError:
    _USER_AGENT = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36"


class Suggestions:
    """Class for YouTube search suggestions and autocomplete."""

    AUTOCOMPLETE_URL = "https://suggestqueries.google.com/complete/search"

    @staticmethod
    def autocomplete(query: str, language: str = "en") -> List[str]:
        """
        Get YouTube autocomplete suggestions for a search query.

        Args:
            query: Search query to get suggestions for
            language: Language code (e.g., 'en', 'es', 'fr')

        Returns:
            List of autocomplete suggestions
        """
        if not query:
            return []

        url = f"{Suggestions.AUTOCOMPLETE_URL}?client=youtube&ds=yt&q={quote(query)}&hl={language}"

        headers = {
            "User-Agent": _USER_AGENT,
            "Accept": "application/json"
        }

        try:
            req = Request(url, headers=headers)
            response = urlopen(req, timeout=10)
            data = response.read().decode('utf-8')

            # Response is JSONP, extract JSON part
            # Format: window.google.ac.h(["query",[["suggestion1"],["suggestion2"],...]])
            start = data.find('(')
            end = data.rfind(')')
            if start != -1 and end != -1:
                json_str = data[start + 1:end]
                parsed = json.loads(json_str)
                if len(parsed) > 1 and isinstance(parsed[1], list):
                    return [item[0] for item in parsed[1] if isinstance(item, list) and item]
            return []
        except Exception:
            return []

    @staticmethod
    def trending_searches(language: str = "en", country: str = "US") -> List[str]:
        """
        Get trending YouTube searches.

        Args:
            language: Language code
            country: Country code

        Returns:
            List of trending search terms
        """
        # Get suggestions for empty-ish queries that return trending
        trending = []
        for seed in ["", "how to", "what is", "best"]:
            suggestions = Suggestions.autocomplete(seed, language)
            trending.extend(suggestions[:3])

        # Remove duplicates while preserving order
        seen = set()
        unique = []
        for item in trending:
            if item not in seen:
                seen.add(item)
                unique.append(item)

        return unique[:20]


if __name__ == "__main__":
    print("Testing autocomplete:")
    suggestions = Suggestions.autocomplete("python tutorial")
    for s in suggestions[:5]:
        print(f"  - {s}")

    print("\nTrending searches:")
    trending = Suggestions.trending_searches()
    for t in trending[:5]:
        print(f"  - {t}")



"""Bing images search."""

from __future__ import annotations

from time import sleep
from typing import List, Optional
from urllib.parse import urlencode

from webscout.scout import Scout
from webscout.search.results import ImagesResult

from .base import BingBase


class BingImagesSearch(BingBase):
    name = "bing"
    category = "images"
    def run(self, *args, **kwargs) -> List[ImagesResult]:
        keywords = args[0] if args else kwargs.get("keywords")
        args[1] if len(args) > 1 else kwargs.get("region", "us")
        safesearch = args[2] if len(args) > 2 else kwargs.get("safesearch", "moderate")
        max_results = args[3] if len(args) > 3 else kwargs.get("max_results", 10)

        if max_results is None:
            max_results = 10

        if not keywords:
            raise ValueError("Keywords are mandatory")

        safe_map = {
            "on": "Strict",
            "moderate": "Moderate",
            "off": "Off"
        }
        safe_map.get(safesearch.lower(), "Moderate")

        # Bing images URL
        url = f"{self.base_url}/images/async"
        params = {
            'q': keywords,
            'first': '1',
            'count': '35',  # Fetch more to get max_results
            'cw': '1177',
            'ch': '759',
            'tsc': 'ImageHoverTitle',
            'layout': 'RowBased_Landscape',
            't': '0',
            'IG': '',
            'SFX': '0',
            'iid': 'images.1'
        }

        results = []
        first = 1
        sfx = 0

        while len(results) < max_results:
            params['first'] = str(first)
            params['SFX'] = str(sfx)
            full_url = f"{url}?{urlencode(params)}"

            try:
                response = self.session.get(full_url, timeout=self.timeout)
                response.raise_for_status()
                html = response.text
            except Exception as e:
                raise Exception(f"Failed to fetch images: {str(e)}")

            soup = Scout(html)
            img_tags = soup.select('a.iusc img')

            for img in img_tags:
                if len(results) >= max_results:
                    break

                title = img.get('alt', '')
                src = img.get('src', '')
                m_attr = img.parent.get('m', '') if img.parent else ''

                # Parse m attribute for full image URL
                image_url = src
                thumbnail = src
                if m_attr:
                    try:
                        import json
                        m_data = json.loads(m_attr)
                        image_url = m_data.get('murl', src)
                        thumbnail = m_data.get('turl', src)
                    except Exception:
                        pass

                source = ''
                if img.parent and img.parent.parent:
                    source_tag = img.parent.parent.select_one('.iusc .lnk')
                    if source_tag:
                        source = source_tag.get_text(strip=True)

                results.append(ImagesResult(
                    title=title,
                    image=image_url,
                    thumbnail=thumbnail,
                    url=image_url,
                    height=0,
                    width=0,
                    source=source
                ))

            first += 35
            sfx += 1

            if self.sleep_interval:
                sleep(self.sleep_interval)

        return results[:max_results]



from typing import Any, Dict, List, Optional

from .utils import request


class Organization:
    """Class for interacting with GitHub organization data without authentication"""

    def __init__(self, org: str):
        """
        Initialize organization client.

        Args:
            org: Organization login name
        """
        if not org:
            raise ValueError("Organization name is required")
        if not isinstance(org, str):
            raise ValueError("Organization name must be a string")

        self.org = org.strip()
        self.base_url = f"https://api.github.com/orgs/{self.org}"

    def get_info(self) -> Dict[str, Any]:
        """
        Get organization information.

        Returns:
            Organization details including name, description, location, etc.
        """
        return request(self.base_url)

    def get_repos(
        self,
        repo_type: str = "all",
        sort: str = "created",
        direction: str = "desc",
        page: int = 1,
        per_page: int = 30
    ) -> List[Dict[str, Any]]:
        """
        List organization repositories.

        Args:
            repo_type: Type of repos (all, public, private, forks, sources, member)
            sort: Sort by (created, updated, pushed, full_name)
            direction: Sort direction (asc, desc)
            page: Page number
            per_page: Results per page (max 100)

        Returns:
            List of organization repositories
        """
        url = f"{self.base_url}/repos?type={repo_type}&sort={sort}&direction={direction}&page={page}&per_page={per_page}"
        return request(url)

    def get_public_members(
        self,
        page: int = 1,
        per_page: int = 30
    ) -> List[Dict[str, Any]]:
        """
        List public members of the organization.

        Args:
            page: Page number
            per_page: Results per page (max 100)

        Returns:
            List of public organization members
        """
        url = f"{self.base_url}/public_members?page={page}&per_page={per_page}"
        return request(url)

    def get_events(
        self,
        page: int = 1,
        per_page: int = 30
    ) -> List[Dict[str, Any]]:
        """
        List public organization events.

        Args:
            page: Page number
            per_page: Results per page (max 100)

        Returns:
            List of public events
        """
        url = f"{self.base_url}/events?page={page}&per_page={per_page}"
        return request(url)



import base64
import json
import random
import re
import string
import time
import uuid
from datetime import datetime
from typing import Any, Dict, Generator, Optional, Union, cast

from curl_cffi import CurlError
from curl_cffi.requests import Session

from webscout import exceptions
from webscout.AIbase import Provider, Response
from webscout.AIutel import (
    AwesomePrompts,
    Conversation,
    Optimizers,
    sanitize_stream,
)


class Toolbaz(Provider):
    """
    A class to interact with the Toolbaz API. Supports streaming responses.
    """

    required_auth = False
    AVAILABLE_MODELS = [
        "gemini-2.5-flash",
        "gemini-2.5-pro",
        "gemini-2.0-flash-thinking",
        "gemini-2.0-flash",
        "claude-sonnet-4",
        "gpt-5",
        "gpt-5.2",
        "gpt-oss-120b",
        "o3-mini",
        "gpt-4o-latest",
        "grok-4-fast",
        "grok-4.1-fast",
        "toolbaz-v4.5-fast",
        "toolbaz_v4",
        "toolbaz_v3.5_pro",
        "deepseek-r1",
        "deepseek-v3.1",
        "deepseek-v3",
        "Llama-4-Maverick",
        "Llama-3.3-70B",
        "mixtral_8x22b",
        "L3-70B-Euryale-v2.1",
        "midnight-rose",
        "unfiltered_x",
    ]

    def __init__(
        self,
        is_conversation: bool = True,
        max_tokens: int = 600,
        timeout: int = 30,
        intro: Optional[str] = None,
        filepath: Optional[str] = None,
        update_file: bool = True,
        proxies: dict = {},
        history_offset: int = 10250,
        act: Optional[str] = None,
        model: str = "gemini-2.0-flash",
        system_prompt: str = "You are a helpful AI assistant.",
    ):
        """
        Initializes the Toolbaz API with given parameters.
        """
        if model not in self.AVAILABLE_MODELS:
            raise ValueError(f"Invalid model: {model}. Choose from: {self.AVAILABLE_MODELS}")

        self.session = Session()
        self.is_conversation = is_conversation
        self.max_tokens_to_sample = max_tokens
        self.timeout = timeout
        self.last_response = {}
        self.system_prompt = system_prompt
        self.model = model
        self.proxies = proxies

        # Set up headers for the curl_cffi session
        self.session.headers.update(
            {
                "user-agent": "Mozilla/5.0 (Linux; Android 10)",
                "accept": "*/*",
                "accept-language": "en-US",
                "cache-control": "no-cache",
                "content-type": "application/x-www-form-urlencoded; charset=UTF-8",
                "origin": "https://toolbaz.com",
                "pragma": "no-cache",
                "referer": "https://toolbaz.com/",
                "sec-fetch-mode": "cors",
            }
        )

        if proxies:
            self.session.proxies.update(cast(Any, proxies))

        # Initialize conversation history
        self.__available_optimizers = (
            method
            for method in dir(Optimizers)
            if callable(getattr(Optimizers, method)) and not method.startswith("__")
        )

        self.conversation = Conversation(
            is_conversation, self.max_tokens_to_sample, filepath, update_file
        )
        act_prompt = (
            AwesomePrompts().get_act(
                cast(Union[str, int], act), default=None, case_insensitive=True
            )
            if act
            else intro
        )
        if act_prompt:
            self.conversation.intro = act_prompt
        self.conversation.history_offset = history_offset

    def random_string(self, length: int) -> str:
        return "".join(random.choices(string.ascii_letters + string.digits, k=length))

    def generate_token(self) -> str:
        payload = {
            "bR6wF": {
                "nV5kP": "Mozilla/5.0 (Linux; Android 10)",
                "lQ9jX": "en-US",
                "sD2zR": "431x958",
                "tY4hL": time.tzname[0] if time.tzname else "UTC",
                "pL8mC": "Linux armv81",
                "cQ3vD": datetime.now().year,
                "hK7jN": datetime.now().hour,
            },
            "uT4bX": {"mM9wZ": [], "kP8jY": []},
            "tuTcS": int(time.time()),
            "tDfxy": None,
            "RtyJt": str(uuid.uuid4()),
        }
        return "d8TW0v" + base64.b64encode(json.dumps(payload).encode()).decode()

    def get_auth(self) -> Dict[str, str]:
        resp = None
        try:
            session_id = self.random_string(36)
            token = self.generate_token()
            data = {"session_id": session_id, "token": token}
            resp = self.session.post(
                "https://data.toolbaz.com/token.php",
                data=data,
            )
            resp.raise_for_status()
            result = resp.json()
            if result.get("success"):
                return {"token": result["token"], "session_id": session_id}

            raise exceptions.FailedToGenerateResponseError(
                f"Authentication failed: API response indicates failure. Response: {result}"
            )
        except CurlError as e:
            raise exceptions.FailedToGenerateResponseError(
                f"Authentication failed due to network error (CurlError): {e}"
            ) from e
        except json.JSONDecodeError as e:
            raise exceptions.FailedToGenerateResponseError(
                f"Authentication failed: Could not decode JSON response. Error: {e}. Response text: {getattr(resp, 'text', 'N/A')}"
            ) from e
        except Exception as e:
            err_text = ""
            if hasattr(e, "response"):
                response_obj = getattr(e, "response")
                if hasattr(response_obj, "text"):
                    err_text = getattr(response_obj, "text")
            raise exceptions.FailedToGenerateResponseError(
                f"Authentication failed due to an unexpected error ({type(e).__name__}): {e} - {err_text}"
            ) from e

    def ask(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False,
        optimizer: Optional[str] = None,
        conversationally: bool = False,
        **kwargs: Any,
    ) -> Response:
        """Sends a prompt to the Toolbaz API and returns the response."""
        if optimizer and optimizer not in self.__available_optimizers:
            raise exceptions.FailedToGenerateResponseError(
                f"Optimizer is not one of {self.__available_optimizers}"
            )

        conversation_prompt = self.conversation.gen_complete_prompt(prompt)
        if optimizer:
            conversation_prompt = getattr(Optimizers, optimizer)(
                conversation_prompt if conversationally else prompt
            )

        auth = self.get_auth()

        data = {
            "text": conversation_prompt,
            "capcha": auth["token"],
            "model": self.model,
            "session_id": auth["session_id"],
        }

        def for_stream():
            try:
                resp = self.session.post(
                    "https://data.toolbaz.com/writing.php",
                    data=data,
                    stream=True,
                    timeout=self.timeout,
                    impersonate="chrome110",
                )
                resp.raise_for_status()

                streaming_text = ""

                processed_stream = sanitize_stream(
                    data=resp.iter_content(chunk_size=None),
                    intro_value=None,
                    to_json=False,
                    skip_regexes=[r"\[model:.*?\]"],
                    yield_raw_on_error=True,
                    raw=raw,
                )

                for content_chunk in processed_stream:
                    if isinstance(content_chunk, bytes):
                        content_chunk = content_chunk.decode("utf-8", errors="ignore")
                    if content_chunk is None:
                        continue
                    if raw:
                        yield content_chunk
                    else:
                        if content_chunk and isinstance(content_chunk, str):
                            streaming_text += content_chunk
                            yield {"text": content_chunk}

                self.last_response = {"text": streaming_text}
                self.conversation.update_chat_history(prompt, streaming_text)

            except CurlError as e:
                raise exceptions.FailedToGenerateResponseError(
                    f"Network error (CurlError): {str(e)}"
                ) from e
            except Exception as e:
                raise exceptions.FailedToGenerateResponseError(
                    f"Unexpected error during stream: {str(e)}"
                ) from e

        def for_non_stream():
            try:
                resp = self.session.post(
                    "https://data.toolbaz.com/writing.php",
                    data=data,
                    timeout=self.timeout,
                    impersonate="chrome110",
                )
                resp.raise_for_status()

                text = resp.text
                text = re.sub(r"\[model:.*?\]", "", text)

                self.last_response = {"text": text}
                self.conversation.update_chat_history(prompt, text)

                return self.last_response

            except CurlError as e:
                raise exceptions.FailedToGenerateResponseError(
                    f"Network error (CurlError): {str(e)}"
                ) from e
            except Exception as e:
                raise exceptions.FailedToGenerateResponseError(f"Unexpected error: {str(e)}") from e

        return for_stream() if stream else for_non_stream()

    def chat(
        self,
        prompt: str,
        stream: bool = False,
        optimizer: Optional[str] = None,
        conversationally: bool = False,
        **kwargs: Any,
    ) -> Union[str, Generator[str, None, None]]:
        """Generates a response from the Toolbaz API."""
        raw = kwargs.get("raw", False)

        def for_stream_chat():
            for response in self.ask(
                prompt, stream=True, raw=raw, optimizer=optimizer, conversationally=conversationally
            ):
                if raw:
                    yield cast(str, response)
                else:
                    yield self.get_message(cast(Response, response))

        def for_non_stream_chat():
            response_dict = self.ask(
                prompt,
                stream=False,
                raw=raw,
                optimizer=optimizer,
                conversationally=conversationally,
            )
            if raw:
                return cast(str, response_dict)
            else:
                return self.get_message(response_dict)

        return for_stream_chat() if stream else for_non_stream_chat()

    def get_message(self, response: Response) -> str:
        if not isinstance(response, dict):
            return str(response)
        response_dict = cast(Dict[str, Any], response)
        return response_dict.get("text", "")


if __name__ == "__main__":
    from rich import print

    print("-" * 80)
    print(f"{'Model':<50} {'Status':<10} {'Response'}")
    print("-" * 80)

    for model in Toolbaz.AVAILABLE_MODELS:
        try:
            test_ai = Toolbaz(model=model, timeout=60)
            response_stream = test_ai.chat("Say 'Hello' in one word", stream=True)
            response_text = ""

            if hasattr(response_stream, "__iter__") and not isinstance(
                response_stream, (str, bytes)
            ):
                for chunk in response_stream:
                    response_text += chunk
            else:
                response_text = str(response_stream)

            if response_text and len(response_text.strip()) > 0:
                status = "✓"
                clean_text = response_text.strip()
                display_text = clean_text[:50] + "..." if len(clean_text) > 50 else clean_text
            else:
                status = "✗ (Stream)"
                display_text = "Empty or invalid stream response"

            print(f"\r{model:<50} {status:<10} {display_text}")

        except Exception as e:
            print(f"\r{model:<50} {'✗':<10} Error: {str(e)}")



import base64
import hashlib
import json
import random
import time
import uuid
from datetime import datetime, timezone
from typing import Any, Dict, Generator, List, Optional, Union, cast

import requests

# Import base classes and utility structures
from webscout.Provider.Openai_comp.base import (
    BaseChat,
    BaseCompletions,
    OpenAICompatibleProvider,
    SimpleModelList,
    Tool,
)
from webscout.Provider.Openai_comp.utils import (
    ChatCompletion,
    ChatCompletionChunk,
    ChatCompletionMessage,
    Choice,
    ChoiceDelta,
    CompletionUsage,
    count_tokens,
)

# ANSI escape codes for formatting
BOLD = "\033[1m"
RED = "\033[91m"
RESET = "\033[0m"


class ChatGPTReversed:
    AVAILABLE_MODELS = [
        "auto",
        "gpt-5-1",
        "gpt-5-1-instant",
        "gpt-5-1-thinking",
        "gpt-5",
        "gpt-5-instant",
        "gpt-5-thinking",
        "gpt-4",
        "gpt-4.1",
        "gpt-4-1",
        "gpt-4.1-mini",
        "gpt-4-1-mini",
        "gpt-4.5",
        "gpt-4-5",
        "gpt-4o",
        "gpt-4o-mini",
        "o1",
        "o1-mini",
        "o3-mini",
        "o3-mini-high",
        "o4-mini",
        "o4-mini-high",
    ]
    csrf_token = None
    initialized = False

    _instance = None

    def __new__(cls, model="auto"):
        if cls._instance is None:
            cls._instance = super(ChatGPTReversed, cls).__new__(cls)
            cls._instance.initialized = False
        return cls._instance

    def __init__(self, model="auto"):
        if self.initialized:
            # Already initialized, just update model if needed
            if model not in self.AVAILABLE_MODELS:
                raise ValueError(f"Invalid model: {model}. Choose from: {self.AVAILABLE_MODELS}")
            self.model = model
            return

        if model not in self.AVAILABLE_MODELS:
            raise ValueError(f"Invalid model: {model}. Choose from: {self.AVAILABLE_MODELS}")

        self.model = model
        self.initialize()

    def initialize(self):
        ChatGPTReversed.initialized = True

    def random_ip(self):
        """Generate a random IP address."""
        return ".".join(str(random.randint(0, 255)) for _ in range(4))

    def random_uuid(self):
        """Generate a random UUID."""
        return str(uuid.uuid4())

    def random_float(self, min_val, max_val):
        """Generate a random float between min and max."""
        return round(random.uniform(min_val, max_val), 4)

    def simulate_bypass_headers(self, accept, spoof_address=False, pre_oai_uuid=None):
        """Simulate browser headers to bypass detection."""
        simulated = {
            "agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/132.0.0.0 Safari/537.36",
            "platform": "Windows",
            "mobile": "?0",
            "ua": 'Not A(Brand";v="8", "Chromium";v="132", "Google Chrome";v="132',
        }

        ip = self.random_ip()
        uuid_val = pre_oai_uuid or self.random_uuid()

        headers = {
            "accept": accept,
            "Content-Type": "application/json",
            "cache-control": "no-cache",
            "Referer": "https://chatgpt.com/",
            "Referrer-Policy": "strict-origin-when-cross-origin",
            "oai-device-id": uuid_val,
            "User-Agent": simulated["agent"],
            "pragma": "no-cache",
            "priority": "u=1, i",
            "sec-ch-ua": f"{simulated['ua']}",
            "sec-ch-ua-mobile": simulated["mobile"],
            "sec-ch-ua-platform": f"{simulated['platform']}",
            "sec-fetch-site": "same-origin",
            "sec-fetch-mode": "cors",
        }

        if spoof_address:
            headers.update(
                {
                    "X-Forwarded-For": ip,
                    "X-Originating-IP": ip,
                    "X-Remote-IP": ip,
                    "X-Remote-Addr": ip,
                    "X-Host": ip,
                    "X-Forwarded-Host": ip,
                }
            )

        return headers

    def generate_proof_token(self, seed: str, difficulty: str, user_agent: Optional[str] = None):
        """
        Improved proof-of-work implementation based on gpt4free/g4f/Provider/openai/proofofwork.py

        Args:
            seed: The seed string for the challenge
            difficulty: The difficulty hex string
            user_agent: Optional user agent string

        Returns:
            The proof token starting with 'gAAAAAB'
        """
        if user_agent is None:
            user_agent = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/132.0.0.0 Safari/537.36"

        screen = random.choice([3008, 4010, 6000]) * random.choice([1, 2, 4])

        # Get current UTC time
        now_utc = datetime.now(timezone.utc)
        parse_time = now_utc.strftime("%a, %d %b %Y %H:%M:%S GMT")

        proof_token = [
            screen,
            parse_time,
            None,
            0,
            user_agent,
            "https://tcr9i.chat.openai.com/v2/35536E1E-65B4-4D96-9D97-6ADB7EFF8147/api.js",
            "dpl=1440a687921de39ff5ee56b92807faaadce73f13",
            "en",
            "en-US",
            None,
            "plugins−[object PluginArray]",
            random.choice(
                [
                    "_reactListeningcfilawjnerp",
                    "_reactListening9ne2dfo1i47",
                    "_reactListening410nzwhan2a",
                ]
            ),
            random.choice(["alert", "ontransitionend", "onprogress"]),
        ]

        diff_len = len(difficulty)
        for i in range(100000):
            proof_token[3] = i
            json_data = json.dumps(proof_token)
            base = base64.b64encode(json_data.encode()).decode()
            hash_value = hashlib.sha3_512((seed + base).encode()).digest()

            if hash_value.hex()[:diff_len] <= difficulty:
                return "gAAAAAB" + base

        # Fallback
        fallback_base = base64.b64encode(f'"{seed}"'.encode()).decode()
        return "gAAAAABwQ8Lk5FbGpA2NcR9dShT6gYjU7VxZ4D" + fallback_base

    def solve_sentinel_challenge(self, seed, difficulty):
        """Solve the sentinel challenge for authentication using improved algorithm."""
        return self.generate_proof_token(seed, difficulty)

    def generate_fake_sentinel_token(self):
        """Generate a fake sentinel token for initial authentication."""
        prefix = "gAAAAAC"

        # More realistic screen sizes
        screen = random.choice([3008, 4010, 6000]) * random.choice([1, 2, 4])

        # Get current UTC time
        now_utc = datetime.now(timezone.utc)
        parse_time = now_utc.strftime("%a, %d %b %Y %H:%M:%S GMT")

        config = [
            screen,
            parse_time,
            4294705152,
            0,
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/132.0.0.0 Safari/537.36",
            "de",
            "de",
            401,
            "mediaSession",
            "location",
            "scrollX",
            self.random_float(1000, 5000),
            str(uuid.uuid4()),
            "",
            12,
            int(time.time() * 1000),
        ]

        base64_str = base64.b64encode(json.dumps(config).encode()).decode()
        return prefix + base64_str

    def parse_response(self, input_text):
        """Parse the response from ChatGPT.

        Args:
            input_text (str): The response text from ChatGPT.

        Returns:
            The complete response as a string.
        """
        parts = [part.strip() for part in input_text.split("\n") if part.strip()]

        for part in parts:
            try:
                if part.startswith("data: "):
                    json_data = json.loads(part[6:])
                    if (
                        json_data.get("message")
                        and json_data["message"].get("status") == "finished_successfully"
                        and json_data["message"].get("metadata", {}).get("is_complete")
                    ):
                        return json_data["message"]["content"]["parts"][0]
            except Exception:
                pass

        return input_text  # Return raw text if parsing fails or no complete message found

    def rotate_session_data(self):
        """Rotate session data to maintain fresh authentication."""
        uuid_val = self.random_uuid()
        csrf_token = self.get_csrf_token(uuid_val)
        sentinel_token = self.get_sentinel_token(uuid_val, csrf_token)

        ChatGPTReversed.csrf_token = csrf_token

        return {"uuid": uuid_val, "csrf": csrf_token, "sentinel": sentinel_token}

    def get_csrf_token(self, uuid_val):
        """Get CSRF token for authentication."""
        if ChatGPTReversed.csrf_token is not None:
            return ChatGPTReversed.csrf_token

        headers = self.simulate_bypass_headers(
            accept="application/json", spoof_address=True, pre_oai_uuid=uuid_val
        )

        response = requests.get("https://chatgpt.com/api/auth/csrf", headers=headers)

        data = response.json()
        if "csrfToken" not in data:
            raise Exception("Failed to fetch required CSRF token")

        return data["csrfToken"]

    def get_sentinel_token(self, uuid_val, csrf):
        """Get sentinel token for authentication."""
        headers = self.simulate_bypass_headers(
            accept="application/json", spoof_address=True, pre_oai_uuid=uuid_val
        )

        test = self.generate_fake_sentinel_token()

        response = requests.post(
            "https://chatgpt.com/backend-anon/sentinel/chat-requirements",
            json={"p": test},
            headers={
                **headers,
                "Cookie": f"__Host-next-auth.csrf-token={csrf}; oai-did={uuid_val}; oai-nav-state=1;",
            },
        )

        data = response.json()
        if "token" not in data or "proofofwork" not in data:
            raise Exception("Failed to fetch required sentinel token")

        oai_sc = None
        for cookie in response.cookies:
            if cookie.name == "oai-sc":
                oai_sc = cookie.value
                break

        if not oai_sc:
            raise Exception("Failed to fetch required oai-sc token")

        challenge_token = self.solve_sentinel_challenge(
            data["proofofwork"]["seed"], data["proofofwork"]["difficulty"]
        )

        return {"token": data["token"], "proof": challenge_token, "oaiSc": oai_sc}

    def complete(self, message, model=None):
        """Complete a message using ChatGPT.

        Args:
            message (str): The message to send to ChatGPT.
            model (str, optional): The model to use. If None, uses the model specified during initialization.
                                   Defaults to None.

        Returns:
            The complete response as a string.
        """
        if not ChatGPTReversed.initialized:
            raise Exception(
                "ChatGPTReversed has not been initialized. Please initialize the instance before calling this method."
            )

        # Use the provided model or fall back to the instance model
        selected_model = model if model else self.model

        # Validate the model
        if selected_model not in self.AVAILABLE_MODELS:
            raise ValueError(
                f"Invalid model: {selected_model}. Choose from: {self.AVAILABLE_MODELS}"
            )

        session_data = self.rotate_session_data()

        headers = self.simulate_bypass_headers(
            accept="plain/text",  # Changed accept header as we expect full response now
            spoof_address=True,
            pre_oai_uuid=session_data["uuid"],
        )

        headers.update(
            {
                "Cookie": f"__Host-next-auth.csrf-token={session_data['csrf']}; oai-did={session_data['uuid']}; oai-nav-state=1; oai-sc={session_data['sentinel']['oaiSc']};",
                "openai-sentinel-chat-requirements-token": session_data["sentinel"]["token"],
                "openai-sentinel-proof-token": session_data["sentinel"]["proof"],
            }
        )

        payload = {
            "action": "next",
            "messages": [
                {
                    "id": self.random_uuid(),
                    "author": {"role": "user"},
                    "content": {"content_type": "text", "parts": [message]},
                    "metadata": {},
                }
            ],
            "parent_message_id": self.random_uuid(),
            "model": selected_model,  # Use the selected model
            "timezone_offset_min": -120,
            "suggestions": [],
            "history_and_training_disabled": False,
            "conversation_mode": {
                "kind": "primary_assistant",
                "plugin_ids": None,  # Ensure web search is not used
            },
            "force_paragen": False,
            "force_paragen_model_slug": "",
            "force_nulligen": False,
            "force_rate_limit": False,
            "reset_rate_limits": False,
            "websocket_request_id": self.random_uuid(),
            "force_use_sse": True,  # Keep SSE for receiving the full response
        }

        response = requests.post(
            "https://chatgpt.com/backend-anon/conversation", json=payload, headers=headers
        )

        if response.status_code != 200:
            raise Exception(f"HTTP error! status: {response.status_code}")

        return self.parse_response(response.text)


class Completions(BaseCompletions):
    def __init__(self, client: "ChatGPT"):
        self._client = client
        self._chatgpt_reversed = None

    def create(
        self,
        *,
        model: str,
        messages: List[Dict[str, str]],
        max_tokens: Optional[int] = None,
        stream: bool = False,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        tools: Optional[List[Union[Tool, Dict[str, Any]]]] = None,  # Support for tool definitions
        tool_choice: Optional[
            Union[str, Dict[str, Any]]
        ] = None,  # Support for tool_choice parameter
        timeout: Optional[int] = None,
        proxies: Optional[dict] = None,
        **kwargs: Any,
    ) -> Union[ChatCompletion, Generator[ChatCompletionChunk, None, None]]:
        """
        Create a chat completion with ChatGPT API.

        Args:
            model: The model to use (from AVAILABLE_MODELS)
            messages: List of message dictionaries with 'role' and 'content'
            max_tokens: Maximum number of tokens to generate
            stream: Whether to stream the response
            temperature: Sampling temperature (0-1)
            top_p: Nucleus sampling parameter (0-1)
            tools: List of tool definitions available for the model to use
            tool_choice: Control over which tool the model should use
            **kwargs: Additional parameters to pass to the API

        Returns:
            If stream=False, returns a ChatCompletion object
            If stream=True, returns a Generator yielding ChatCompletionChunk objects
        """
        # Initialize ChatGPTReversed if not already initialized
        if self._chatgpt_reversed is None:
            self._chatgpt_reversed = ChatGPTReversed(model=model)

        # Use streaming implementation if requested
        if stream:
            return self._create_streaming(
                model=model,
                messages=messages,
                max_tokens=max_tokens,
                temperature=temperature,
                top_p=top_p,
                tools=tools,
                tool_choice=tool_choice,
                timeout=timeout,
                proxies=proxies,
                **kwargs,
            )

        # Otherwise use non-streaming implementation
        return self._create_non_streaming(
            model=model,
            messages=messages,
            max_tokens=max_tokens,
            temperature=temperature,
            top_p=top_p,
            tools=tools,
            tool_choice=tool_choice,
            timeout=timeout,
            proxies=proxies,
            **kwargs,
        )

    def _create_streaming(
        self,
        *,
        model: str,
        messages: List[Dict[str, str]],
        max_tokens: Optional[int] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        tools: Optional[List[Union[Tool, Dict[str, Any]]]] = None,
        tool_choice: Optional[Union[str, Dict[str, Any]]] = None,
        timeout: Optional[int] = None,
        proxies: Optional[dict] = None,
        **kwargs: Any,
    ) -> Generator[ChatCompletionChunk, None, None]:
        """Implementation for streaming chat completions."""
        try:
            # Generate request ID and timestamp
            request_id = str(uuid.uuid4())
            created_time = int(time.time())

            # Get the last user message
            last_user_message = None
            for msg in reversed(messages):
                if msg["role"] == "user":
                    last_user_message = msg["content"]
                    break

            if not last_user_message:
                raise ValueError("No user message found in the conversation")

            # Initialize ChatGPTReversed if not already initialized
            if self._chatgpt_reversed is None:
                self._chatgpt_reversed = ChatGPTReversed(model=model)

            # Create a proper streaming request to ChatGPT
            session_data = self._chatgpt_reversed.rotate_session_data()

            headers = self._chatgpt_reversed.simulate_bypass_headers(
                accept="text/event-stream", spoof_address=True, pre_oai_uuid=session_data["uuid"]
            )

            headers.update(
                {
                    "Cookie": f"__Host-next-auth.csrf-token={session_data['csrf']}; oai-did={session_data['uuid']}; oai-nav-state=1; oai-sc={session_data['sentinel']['oaiSc']};",
                    "openai-sentinel-chat-requirements-token": session_data["sentinel"]["token"],
                    "openai-sentinel-proof-token": session_data["sentinel"]["proof"],
                }
            )

            # Format messages properly for ChatGPT
            formatted_messages = []
            for i, msg in enumerate(messages):
                formatted_messages.append(
                    {
                        "id": str(uuid.uuid4()),
                        "author": {"role": msg["role"]},
                        "content": {"content_type": "text", "parts": [msg["content"]]},
                        "metadata": {},
                    }
                )

            payload = {
                "action": "next",
                "messages": formatted_messages,
                "parent_message_id": str(uuid.uuid4()),
                "model": model,
                "timezone_offset_min": -120,
                "suggestions": [],
                "history_and_training_disabled": False,
                "conversation_mode": {"kind": "primary_assistant", "plugin_ids": None},
                "force_paragem": False,
                "force_paragem_model_slug": "",
                "force_nulligen": False,
                "force_rate_limit": False,
                "reset_rate_limits": False,
                "websocket_request_id": str(uuid.uuid4()),
                "force_use_sse": True,
            }

            # Add optional parameters if provided
            if max_tokens is not None:
                payload["max_tokens"] = max_tokens
            if temperature is not None:
                payload["temperature"] = temperature
            if top_p is not None:
                payload["top_p"] = top_p

            # Make the actual streaming request
            response = requests.post(
                "https://chatgpt.com/backend-anon/conversation",
                json=payload,
                headers=headers,
                stream=True,
                timeout=timeout or 30,
            )

            response.raise_for_status()

            # Track conversation state
            full_content = ""
            prompt_tokens = count_tokens(str(messages))
            completion_tokens = 0
            total_tokens = prompt_tokens

            # Process the streaming response
            for line in response.iter_lines(decode_unicode=True):
                if line:
                    if line.startswith("data: "):
                        data_str = line[6:]  # Remove "data: " prefix

                        # Handle [DONE] message
                        if data_str.strip() == "[DONE]":
                            # Final chunk with finish_reason
                            delta = ChoiceDelta(content=None)
                            choice = Choice(index=0, delta=delta, finish_reason="stop")
                            chunk = ChatCompletionChunk(
                                id=request_id, choices=[choice], created=created_time, model=model
                            )
                            chunk.usage = {
                                "prompt_tokens": prompt_tokens,
                                "completion_tokens": completion_tokens,
                                "total_tokens": total_tokens,
                            }
                            yield chunk
                            break

                        try:
                            data = json.loads(data_str)

                            # Handle different types of messages
                            if data.get("message"):
                                message = data["message"]

                                # Handle assistant responses
                                if message.get("author", {}).get("role") == "assistant":
                                    content_parts = message.get("content", {}).get("parts", [])
                                    if content_parts:
                                        new_content = content_parts[0]

                                        # Get the delta (new content since last chunk)
                                        delta_content = (
                                            new_content[len(full_content) :]
                                            if new_content.startswith(full_content)
                                            else new_content
                                        )
                                        full_content = new_content
                                        completion_tokens = count_tokens(full_content)
                                        total_tokens = prompt_tokens + completion_tokens

                                        # Only yield chunk if there's new content
                                        if delta_content:
                                            delta = ChoiceDelta(
                                                content=delta_content, role="assistant"
                                            )
                                            choice = Choice(
                                                index=0, delta=delta, finish_reason=None
                                            )
                                            chunk = ChatCompletionChunk(
                                                id=request_id,
                                                choices=[choice],
                                                created=created_time,
                                                model=model,
                                            )
                                            chunk.usage = {
                                                "prompt_tokens": prompt_tokens,
                                                "completion_tokens": completion_tokens,
                                                "total_tokens": total_tokens,
                                            }
                                            yield chunk

                                # Handle finish status
                                if message.get("status") == "finished_successfully":
                                    pass

                            elif data.get("type") == "message_stream_complete":
                                # Stream is complete
                                pass

                        except json.JSONDecodeError:
                            # Skip invalid JSON lines
                            continue

        except Exception as e:
            raise IOError(f"ChatGPT request failed: {e}") from e

    def _create_non_streaming(
        self,
        *,
        model: str,
        messages: List[Dict[str, str]],
        max_tokens: Optional[int] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        tools: Optional[List[Union[Tool, Dict[str, Any]]]] = None,
        tool_choice: Optional[Union[str, Dict[str, Any]]] = None,
        timeout: Optional[int] = None,
        proxies: Optional[dict] = None,
        **kwargs: Any,
    ) -> ChatCompletion:
        """Implementation for non-streaming chat completions."""
        try:
            # Generate request ID and timestamp
            request_id = str(uuid.uuid4())
            created_time = int(time.time())

            # Initialize ChatGPTReversed if not already initialized
            if self._chatgpt_reversed is None:
                self._chatgpt_reversed = ChatGPTReversed(model=model)

            # Create a proper request to ChatGPT
            session_data = self._chatgpt_reversed.rotate_session_data()

            headers = self._chatgpt_reversed.simulate_bypass_headers(
                accept="text/event-stream", spoof_address=True, pre_oai_uuid=session_data["uuid"]
            )

            headers.update(
                {
                    "Cookie": f"__Host-next-auth.csrf-token={session_data['csrf']}; oai-did={session_data['uuid']}; oai-nav-state=1; oai-sc={session_data['sentinel']['oaiSc']};",
                    "openai-sentinel-chat-requirements-token": session_data["sentinel"]["token"],
                    "openai-sentinel-proof-token": session_data["sentinel"]["proof"],
                }
            )

            # Format messages properly for ChatGPT
            formatted_messages = []
            for i, msg in enumerate(messages):
                formatted_messages.append(
                    {
                        "id": str(uuid.uuid4()),
                        "author": {"role": msg["role"]},
                        "content": {"content_type": "text", "parts": [msg["content"]]},
                        "metadata": {},
                    }
                )

            payload = {
                "action": "next",
                "messages": formatted_messages,
                "parent_message_id": str(uuid.uuid4()),
                "model": model,
                "timezone_offset_min": -120,
                "suggestions": [],
                "history_and_training_disabled": False,
                "conversation_mode": {"kind": "primary_assistant", "plugin_ids": None},
                "force_paragem": False,
                "force_paragem_model_slug": "",
                "force_nulligen": False,
                "force_rate_limit": False,
                "reset_rate_limits": False,
                "websocket_request_id": str(uuid.uuid4()),
                "force_use_sse": True,
            }

            # Add optional parameters if provided
            if max_tokens is not None:
                payload["max_tokens"] = max_tokens
            if temperature is not None:
                payload["temperature"] = temperature
            if top_p is not None:
                payload["top_p"] = top_p

            # Make the request and collect full response
            response = requests.post(
                "https://chatgpt.com/backend-anon/conversation",
                json=payload,
                headers=headers,
                stream=True,
                timeout=timeout or 30,
            )

            response.raise_for_status()

            # Collect and parse the full response
            full_response = ""
            for line in response.iter_lines(decode_unicode=True):
                if line:
                    if line.startswith("data: "):
                        data_str = line[6:]  # Remove "data: " prefix

                        # Handle [DONE] message
                        if data_str.strip() == "[DONE]":
                            break

                        try:
                            data = json.loads(data_str)

                            # Handle assistant responses
                            if (
                                data.get("message")
                                and data["message"].get("author", {}).get("role") == "assistant"
                            ):
                                content_parts = data["message"].get("content", {}).get("parts", [])
                                if content_parts:
                                    full_response = content_parts[0]

                        except json.JSONDecodeError:
                            # Skip invalid JSON lines
                            continue

            # Create the completion message
            message = ChatCompletionMessage(role="assistant", content=full_response)

            # Create the choice
            choice = Choice(index=0, message=message, finish_reason="stop")

            # Calculate token usage using count_tokens
            # Count tokens in the input messages (prompt)
            prompt_tokens = count_tokens(str(messages))
            # Count tokens in the response (completion)
            completion_tokens = count_tokens(full_response)
            total_tokens = prompt_tokens + completion_tokens

            usage = CompletionUsage(
                prompt_tokens=prompt_tokens,
                completion_tokens=completion_tokens,
                total_tokens=total_tokens,
            )

            # Create the completion object with correct OpenAI format
            completion = ChatCompletion(
                id=request_id,
                choices=[choice],
                created=created_time,
                model=model,
                usage=usage,
            )

            return completion

        except Exception as e:
            print(f"{RED}Error during ChatGPT non-stream request: {e}{RESET}")
            raise IOError(f"ChatGPT request failed: {e}") from e


class Chat(BaseChat):
    def __init__(self, client: "ChatGPT"):
        self.completions = Completions(client)


class ChatGPT(OpenAICompatibleProvider):
    """
    OpenAI-compatible client for ChatGPT API.

    Usage:
        client = ChatGPT()
        response = client.chat.completions.create(
            model="auto",
            messages=[{"role": "user", "content": "Hello!"}]
        )
        print(response.choices[0].message.content)
    """

    required_auth = False

    def __init__(
        self,
        api_key: Optional[str] = None,
        tools: Optional[List[Tool]] = None,
        proxies: Optional[Dict[str, str]] = None,
    ):
        """
        Initialize the ChatGPT client.

        Args:
            api_key: Optional API key (not used by ChatGPTReversed but included for interface compatibility)
            tools: Optional list of tools to register with the provider
            proxies: Optional proxy configuration dict, e.g. {"http": "http://proxy:8080", "https": "https://proxy:8080"}
        """
        super().__init__(api_key=api_key, tools=tools, proxies=proxies)
        # Initialize chat interface
        self.chat = Chat(self)

    @property
    def AVAILABLE_MODELS(self):
        return ChatGPTReversed.AVAILABLE_MODELS

    @property
    def models(self) -> SimpleModelList:
        return SimpleModelList(self.AVAILABLE_MODELS)


if __name__ == "__main__":
    # Example usage
    client = ChatGPT()
    response = client.chat.completions.create(
        model="o4-mini-high", messages=[{"role": "user", "content": "How many r in strawberry"}]
    )
    if isinstance(response, ChatCompletion):
        if response.choices[0].message and response.choices[0].message.content:
            print(response.choices[0].message.content)
    print()



import json
from typing import Any, Dict, Generator, Iterator, List, Optional, Union, cast

from curl_cffi import CurlError
from curl_cffi.requests import Session

from webscout import exceptions
from webscout.AIbase import Provider, Response
from webscout.AIutel import AwesomePrompts, Conversation, Optimizers
from webscout.litagent import LitAgent


class DeepAI(Provider):
    """
    DeepAI Chat Provider

    A provider for DeepAI's chat functionality, supporting both streaming and non-streaming responses.
    Structured similarly to other providers like DeepInfra and X0GPT.
    """
    required_auth = True
    AVAILABLE_MODELS = [
        "standard",
        "genius",
        "online",
        "supergenius",
        "onlinegenius",
        "deepseek-v3.2",
        "gemini-2.5-flash-lite",
        "qwen3-30b-a3b",
        "gpt-5-nano",
        "gpt-oss-120b",
        "gpt-5-chat-latest",
        "claude-opus-4-1",
        "llama-4-scout",
        "claude-4.5-sonnet",
        "deepseek-v3.1-terminus",
        "llama-3.3-70b-instruct",
        "grok-4",
        "claude-sonnet-4",
        "qwen3-coder",
        "gpt-5",
        "kimi-k2-0905",
        "claude-opus-4",
        "gpt-5-mini",
        "gemini-2.5-pro",
        "grok-code-fast-1",
        "gpt-4.1",

    ]

    def __init__(
        self,
        api_key: str = "tryit-53926507126-2c8a2543c7b5638ca6b92b6e53ef2d2b",
        timeout: int = 30,
        proxies: Optional[Dict[str, str]] = None,
        model: str = "standard",
        chat_style: str = "chat",
        enabled_tools: Optional[List[str]] = None,
        is_conversation: bool = True,
        max_tokens: int = 2048,
        intro: Optional[str] = None,
        filepath: Optional[str] = None,
        update_file: bool = True,
        history_offset: int = 10250,
        act: Optional[str] = None,
        system_prompt: str = "You are a helpful assistant.",
        browser: str = "chrome",
        **kwargs
    ):
        """
        Initialize the DeepAI provider.

        Args:
            api_key: API key for authentication (trial key provided by default)
            timeout: Request timeout in seconds
            proxies: Proxy configuration
            model: Model to use (default: "standard")
            chat_style: Chat style (default: "chat")
            enabled_tools: List of enabled tools (default: ["image_generator"])
            is_conversation: Whether to maintain conversation history
            max_tokens: Maximum tokens for conversation
            intro: Introduction prompt
            filepath: Path to conversation history file
            update_file: Whether to update history file
            history_offset: History offset for truncation
            act: Act prompt from AwesomePrompts
            system_prompt: System prompt for the AI
            browser: Browser type for fingerprinting
            **kwargs: Additional arguments
        """
        if model not in self.AVAILABLE_MODELS:
            raise ValueError(f"Invalid model: {model}. Choose from: {self.AVAILABLE_MODELS}")

        self.url = "https://api.deepai.org/hacking_is_a_serious_crime"
        self.api_key = api_key
        self.proxies = proxies or {}
        self.model = model
        self.chat_style = chat_style
        self.enabled_tools = enabled_tools or ["image_generator"]
        self.system_prompt = system_prompt
        self.is_conversation = is_conversation
        self.max_tokens_to_sample = max_tokens
        self.timeout = timeout
        self.last_response = {}

        # LitAgent for fingerprinting
        self.agent = LitAgent()
        self.fingerprint = self.agent.generate_fingerprint(browser)

        # Setup headers similar to other providers
        self.headers = {
            "User-Agent": self.fingerprint.get("user_agent", "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36"),
            "Accept": "*/*",
            "Accept-Language": self.fingerprint.get("accept_language", "en-US,en;q=0.9"),
            "Origin": "https://deepai.org",
            "Referer": "https://deepai.org/",
            "Sec-CH-UA": self.fingerprint.get("sec_ch_ua", '"Not_A Brand";v="8", "Chromium";v="120", "Google Chrome";v="120"'),
            "Sec-CH-UA-Mobile": "?0",
            "Sec-CH-UA-Platform": f'"{self.fingerprint.get("platform", "Windows")}"',
            "api-key": self.api_key
        }

        # Setup session
        self.session = Session()
        self.session.headers.update(self.headers)
        if self.proxies:
            self.session.proxies = self.proxies  # type: ignore

        # Optimizers
        self.__available_optimizers = (
            method
            for method in dir(Optimizers)
            if callable(getattr(Optimizers, method)) and not method.startswith("__")
        )

        # Conversation setup similar to other providers
        self.conversation = Conversation(
            is_conversation, self.max_tokens_to_sample, filepath, update_file
        )
        self.conversation.history_offset = history_offset

        act_prompt = (
            AwesomePrompts().get_act(cast(Union[str, int], act), default=None, case_insensitive=True
            )
            if act
            else intro
        )
        if act_prompt:
            self.conversation.intro = act_prompt

    def refresh_identity(self, browser: Optional[str] = None):
        """
        Refreshes the browser identity fingerprint.

        Args:
            browser: Specific browser to use for the new fingerprint
        """
        browser = browser or self.fingerprint.get("browser_type", "chrome")
        self.fingerprint = self.agent.generate_fingerprint(browser)

        # Update relevant headers
        self.headers.update({
            "User-Agent": str(self.fingerprint.get("user_agent", "")),
            "Accept-Language": str(self.fingerprint.get("accept_language", "")),
            "Sec-CH-UA": str(self.fingerprint.get("sec_ch_ua", "")),
            "Sec-CH-UA-Platform": f'"{self.fingerprint.get("platform", "")}"',
        })

        self.session.headers.update(self.headers)
        return self.fingerprint

    def ask(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False,
        optimizer: Optional[str] = None,
        conversationally: bool = False,
        **kwargs: Any,
    ) -> Response:
        """
        Send a prompt to DeepAI and get the response.

        Args:
            prompt: The prompt to send
            stream: Whether to stream the response (fake streaming: yields full response in one chunk)
            raw: Whether to return raw response
            optimizer: Optimizer to use
            conversationally: Whether to apply optimizer to full conversation
            **kwargs: Additional arguments

        Returns:
            Response dictionary with the AI response or generator for streaming
        """
        # Generate conversation prompt similar to other providers
        conversation_prompt = self.conversation.gen_complete_prompt(prompt)

        if optimizer:
            if optimizer in self.__available_optimizers:
                conversation_prompt = getattr(Optimizers, optimizer)(
                    conversation_prompt if conversationally else prompt
                )
            else:
                raise Exception(f"Optimizer is not one of {list(self.__available_optimizers)}")

        # Prepare form data
        # Use conversation_prompt as user content in chatHistory
        chat_history = [
            {"role": "system", "content": self.system_prompt},
            {"role": "user", "content": conversation_prompt}
        ]
        data = {
            "chat_style": self.chat_style,
            "chatHistory": json.dumps(chat_history),
            "model": self.model,
            "hacker_is_stinky": "very_stinky",
            "enabled_tools": json.dumps(self.enabled_tools)
        }

        # Always perform non-streaming request
        try:
            # Make request with curl_cffi
            response = self.session.post(
                self.url,
                data=data,
                timeout=self.timeout,
                impersonate="chrome110"
            )
            response.raise_for_status()

            # Get response text
            result = response.text.strip()

            # Update last response and conversation history
            self.last_response = {"text": result}
            self.conversation.update_chat_history(prompt, result)

            if stream:
                # Fake streaming: yield the full response in one chunk
                if raw:
                    yield result
                else:
                    yield self.last_response
            else:
                return self.last_response if not raw else result

        except CurlError as e:
            raise exceptions.FailedToGenerateResponseError(f"DeepAI API request failed (CurlError): {str(e)}")
        except Exception as e:
            raise exceptions.FailedToGenerateResponseError(f"DeepAI API request failed ({type(e).__name__}): {str(e)}")

    def chat(
        self,
        prompt: str,
        stream: bool = False,
        optimizer: Optional[str] = None,
        conversationally: bool = False,
        raw: bool = False,
        **kwargs: Any,
    ) -> Union[str, Generator[str, None, None]]:
        """
        Send a chat message to DeepAI and get the response.

        Args:
            prompt: The prompt to send
            stream: Whether to stream the response (fake streaming: yields full response in one chunk)
            optimizer: Optimizer to use
            conversationally: Whether to apply optimizer to full conversation
            raw: Whether to return raw response
            **kwargs: Additional arguments

        Returns:
            The AI response as a string or generator for streaming
        """
        if stream:
            for resp in self.ask(
                prompt=prompt,
                stream=True,
                raw=raw,
                optimizer=optimizer,
                conversationally=conversationally,
                **kwargs
            ):
                if raw:
                    yield resp
                else:
                    yield self.get_message(resp)
        else:
            response = self.ask(
                prompt=prompt,
                stream=False,
                raw=raw,
                optimizer=optimizer,
                conversationally=conversationally,
                **kwargs
            )
            if raw:
                return response
            else:
                return self.get_message(response)

    def get_message(self, response: Response) -> str:
        """
        Extract the message from the response.

        Args:
            response: Response obtained from ask method

        Returns:
            The message text
        """
        if isinstance(response, dict):
            return cast(Dict[str, Any], response).get("text", "")
        elif isinstance(response, str):
            return response
        else:
            return str(response)

    @classmethod
    def get_models(cls) -> List[str]:
        """
        Get available models.

        Returns:
            List of available model names
        """
        return cls.AVAILABLE_MODELS

    @classmethod
    def get_chat_styles(cls) -> List[str]:
        """
        Get available chat styles.

        Returns:
            List of available chat styles
        """
        return ["chat"]


if __name__ == "__main__":
    # Test similar to other providers, using stream=True for consistency
    print("-" * 80)
    print(f"{'Model':<50} {'Status':<10} {'Response'}")
    print("-" * 80)

    for model in DeepAI.AVAILABLE_MODELS:
        try:
            test_ai = DeepAI(model=model, timeout=60)
            response = test_ai.chat("Say 'Hello' in one word", stream=True)
            response_text = ""
            if hasattr(response, "__iter__") and not isinstance(response, (str, bytes)):
                for chunk in response:
                    response_text += chunk
            else:
                response_text = str(response)

            if response_text and len(response_text.strip()) > 0:
                status = "✓"
                # Clean and truncate response
                clean_text = response_text.strip().encode('utf-8', errors='ignore').decode('utf-8')
                display_text = clean_text[:50] + "..." if len(clean_text) > 50 else clean_text
            else:
                status = "✗"
                display_text = "Empty or invalid response"
            print(f"{model:<50} {status:<10} {display_text}")
        except Exception as e:
            print(f"{model:<50} {'✗':<10} {str(e)}")
